#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import api as db_api
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log
from oslo_policy import policy as oslo_policy
from oslo_utils import excutils
import stevedore

from gbpservice._i18n import _LE
from gbpservice._i18n import _LI
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy.common import exceptions as gp_exc


LOG = log.getLogger(__name__)
cfg.CONF.import_opt('policy_drivers',
                    'gbpservice.neutron.services.grouppolicy.config',
                    group='group_policy')


class PolicyDriverManager(stevedore.named.NamedExtensionManager):
    """Manage group policy enforcement using drivers.

    Defines precommit and post_commit operations for the
    create, update, and delete operations on the Group Policy
    resources.

    precommit operation:
    Notifies all policy drivers during policy_target creation.

    Raises neutron.services.grouppolicy.common.GroupPolicyDriverError
    if any policy driver create_policy_target_precommit call fails.

    Called within the database transaction. If a policy driver
    raises an exception, then a GroupPolicyDriverError is propogated
    to the caller, triggering a rollback. There is no guarantee
    that all policy drivers are called in this case.

    postcommit operation:
    Notifies all policy drivers after policy_target creation.

    Raises neutron.services.grouppolicy.common.GroupPolicyDriverError
    if any policy driver create_policy_target_postcommit call fails.

    Called after the database transaction. If a policy driver
    raises an exception, then a GroupPolicyDriverError is propagated
    to the caller, where the policy_target will be deleted, triggering
    any required cleanup. There is no guarantee that all policy
    drivers are called in this case.
    """

    def __init__(self):
        # Registered policy drivers, keyed by name.
        self.policy_drivers = {}
        # Ordered list of policy drivers, defining
        # the order in which the drivers are called.
        self.ordered_policy_drivers = []
        self.reverse_ordered_policy_drivers = []

        LOG.info(_LI("Configured policy driver names: %s"),
                 cfg.CONF.group_policy.policy_drivers)
        super(PolicyDriverManager,
              self).__init__('gbpservice.neutron.group_policy.policy_drivers',
                             cfg.CONF.group_policy.policy_drivers,
                             invoke_on_load=True,
                             name_order=True)
        LOG.info(_LI("Loaded policy driver names: %s"), self.names())
        self._register_policy_drivers()

    def _register_policy_drivers(self):
        """Register all policy drivers.

        This method should only be called once in the PolicDriverManager
        constructor.
        """
        for ext in self:
            self.policy_drivers[ext.name] = ext
            self.ordered_policy_drivers.append(ext)

        self.reverse_ordered_policy_drivers = self.ordered_policy_drivers[::-1]
        LOG.info(_LI("Registered policy drivers: %s"),
                 [driver.name for driver in self.ordered_policy_drivers])

    def initialize(self):
        # Group Policy bulk operations requires each driver to support them.
        # However, this is currently not supported at the plugin level,
        # so setting it to False. When the plugin does support it, we can
        # set it to True such that the drivers can override it.
        self.native_bulk_support = False
        for driver in self.ordered_policy_drivers:
            LOG.info(_LI("Initializing policy driver '%s'"), driver.name)
            driver.obj.initialize()
            self.native_bulk_support &= getattr(driver.obj,
                                                'native_bulk_support', True)

    def _call_on_drivers(self, method_name, context,
                         continue_on_failure=False):
        """Helper method for calling a method across all policy drivers.

        :param method_name: name of the method to call
        :param context: context parameter to pass to each method call
        :param continue_on_failure: whether or not to continue to call
        all policy drivers once one has raised an exception
        :raises: neutron.services.group_policy.common.GroupPolicyDriverError
        if any policy driver call fails.
        """
        error = False
        drivers = (self.ordered_policy_drivers if not
                   method_name.startswith('delete') else
                   self.reverse_ordered_policy_drivers)
        for driver in drivers:
            try:
                getattr(driver.obj, method_name)(context)
            except Exception as e:
                if db_api.is_retriable(e):
                    with excutils.save_and_reraise_exception():
                        LOG.debug(
                            "Policy driver '%(name)s' failed in"
                            " %(method)s, operation will be retried",
                            {'name': driver.name, 'method': method_name}
                        )
                elif isinstance(e, gp_exc.GroupPolicyException) or isinstance(
                    e, n_exc.NeutronException) or isinstance(
                        e, oslo_policy.PolicyNotAuthorized):
                    with excutils.save_and_reraise_exception():
                        LOG.exception(
                            _LE("Policy driver '%(name)s' failed in"
                                " %(method)s"),
                            {'name': driver.name, 'method': method_name}
                        )
                else:
                    error = True
                    # We are eating a non-GBP/non-Neutron exception here
                    LOG.exception(
                        _LE("Policy driver '%(name)s' failed in %(method)s"),
                        {'name': driver.name, 'method': method_name})
                    if not continue_on_failure:
                        break
        if error:
            raise gp_exc.GroupPolicyDriverError(method=method_name)

    def ensure_tenant(self, plugin_context, tenant_id):
        for driver in self.ordered_policy_drivers:
            if isinstance(driver.obj, api.PolicyDriver):
                try:
                    driver.obj.ensure_tenant(plugin_context, tenant_id)
                except Exception as e:
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug("Policy driver '%(driver)s' failed in "
                                      "ensure_tenant, operation will "
                                      "be retried", {'driver': driver.name})
                    else:
                        LOG.exception(_LE("Policy driver '%s' failed in "
                                          "ensure_tenant"), driver.name)
                        raise gp_exc.GroupPolicyDriverError(
                                method="ensure_tenant")

    def create_policy_target_precommit(self, context):
        self._call_on_drivers("create_policy_target_precommit", context)

    def create_policy_target_postcommit(self, context):
        self._call_on_drivers("create_policy_target_postcommit", context)

    def update_policy_target_precommit(self, context):
        self._call_on_drivers("update_policy_target_precommit", context)

    def update_policy_target_postcommit(self, context):
        self._call_on_drivers("update_policy_target_postcommit", context)

    def delete_policy_target_precommit(self, context):
        self._call_on_drivers("delete_policy_target_precommit", context)

    def delete_policy_target_postcommit(self, context):
        self._call_on_drivers("delete_policy_target_postcommit", context,
                              continue_on_failure=True)

    def get_policy_target_status(self, context):
        self._call_on_drivers("get_policy_target_status", context)

    def create_policy_target_group_precommit(self, context):
        self._call_on_drivers("create_policy_target_group_precommit", context)

    def create_policy_target_group_postcommit(self, context):
        self._call_on_drivers("create_policy_target_group_postcommit", context)

    def update_policy_target_group_precommit(self, context):
        self._call_on_drivers("update_policy_target_group_precommit", context)

    def update_policy_target_group_postcommit(self, context):
        self._call_on_drivers("update_policy_target_group_postcommit", context)

    def delete_policy_target_group_precommit(self, context):
        self._call_on_drivers("delete_policy_target_group_precommit", context)

    def delete_policy_target_group_postcommit(self, context):
        self._call_on_drivers("delete_policy_target_group_postcommit", context,
                              continue_on_failure=True)

    def get_policy_target_group_status(self, context):
        self._call_on_drivers("get_policy_target_group_status", context)

    def create_application_policy_group_precommit(self, context):
        self._call_on_drivers("create_application_policy_group_precommit",
                              context)

    def create_application_policy_group_postcommit(self, context):
        self._call_on_drivers("create_application_policy_group_postcommit",
                              context)

    def update_application_policy_group_precommit(self, context):
        self._call_on_drivers("update_application_policy_group_precommit",
                              context)

    def update_application_policy_group_postcommit(self, context):
        self._call_on_drivers("update_application_policy_group_postcommit",
                              context)

    def delete_application_policy_group_precommit(self, context):
        self._call_on_drivers("delete_application_policy_group_precommit",
                              context)

    def delete_application_policy_group_postcommit(self, context):
        self._call_on_drivers("delete_application_policy_group_postcommit",
                              context, continue_on_failure=True)

    def get_application_policy_group_status(self, context):
        self._call_on_drivers("get_application_policy_group_status", context)

    def create_l2_policy_precommit(self, context):
        self._call_on_drivers("create_l2_policy_precommit", context)

    def create_l2_policy_postcommit(self, context):
        self._call_on_drivers("create_l2_policy_postcommit", context)

    def update_l2_policy_precommit(self, context):
        self._call_on_drivers("update_l2_policy_precommit", context)

    def update_l2_policy_postcommit(self, context):
        self._call_on_drivers("update_l2_policy_postcommit", context)

    def delete_l2_policy_precommit(self, context):
        self._call_on_drivers("delete_l2_policy_precommit", context)

    def delete_l2_policy_postcommit(self, context):
        self._call_on_drivers("delete_l2_policy_postcommit", context,
                              continue_on_failure=True)

    def get_l2_policy_status(self, context):
        self._call_on_drivers("get_l2_policy_status", context)

    def create_l3_policy_precommit(self, context):
        self._call_on_drivers("create_l3_policy_precommit", context)

    def create_l3_policy_postcommit(self, context):
        self._call_on_drivers("create_l3_policy_postcommit", context)

    def update_l3_policy_precommit(self, context):
        self._call_on_drivers("update_l3_policy_precommit", context)

    def update_l3_policy_postcommit(self, context):
        self._call_on_drivers("update_l3_policy_postcommit", context)

    def delete_l3_policy_precommit(self, context):
        self._call_on_drivers("delete_l3_policy_precommit", context)

    def delete_l3_policy_postcommit(self, context):
        self._call_on_drivers("delete_l3_policy_postcommit", context,
                              continue_on_failure=True)

    def get_l3_policy_status(self, context):
        self._call_on_drivers("get_l3_policy_status", context)

    def create_network_service_policy_precommit(self, context):
        self._call_on_drivers(
            "create_network_service_policy_precommit", context)

    def create_network_service_policy_postcommit(self, context):
        self._call_on_drivers(
            "create_network_service_policy_postcommit", context)

    def update_network_service_policy_precommit(self, context):
        self._call_on_drivers(
            "update_network_service_policy_precommit", context)

    def update_network_service_policy_postcommit(self, context):
        self._call_on_drivers(
            "update_network_service_policy_postcommit", context)

    def delete_network_service_policy_precommit(self, context):
        self._call_on_drivers(
            "delete_network_service_policy_precommit", context)

    def delete_network_service_policy_postcommit(self, context):
        self._call_on_drivers(
            "delete_network_service_policy_postcommit", context,
            continue_on_failure=True)

    def get_network_service_policy_status(self, context):
        self._call_on_drivers("get_network_service_policy_status", context)

    def create_policy_classifier_precommit(self, context):
        self._call_on_drivers("create_policy_classifier_precommit", context)

    def create_policy_classifier_postcommit(self, context):
        self._call_on_drivers("create_policy_classifier_postcommit", context)

    def update_policy_classifier_precommit(self, context):
        self._call_on_drivers("update_policy_classifier_precommit", context)

    def update_policy_classifier_postcommit(self, context):
        self._call_on_drivers("update_policy_classifier_postcommit", context)

    def delete_policy_classifier_precommit(self, context):
        self._call_on_drivers("delete_policy_classifier_precommit", context)

    def delete_policy_classifier_postcommit(self, context):
        self._call_on_drivers("delete_policy_classifier_postcommit", context,
                              continue_on_failure=True)

    def get_policy_classifier_status(self, context):
        self._call_on_drivers("get_policy_classifier_status", context)

    def create_policy_action_precommit(self, context):
        self._call_on_drivers("create_policy_action_precommit", context)

    def create_policy_action_postcommit(self, context):
        self._call_on_drivers("create_policy_action_postcommit", context)

    def update_policy_action_precommit(self, context):
        self._call_on_drivers("update_policy_action_precommit", context)

    def update_policy_action_postcommit(self, context):
        self._call_on_drivers("update_policy_action_postcommit", context)

    def delete_policy_action_precommit(self, context):
        self._call_on_drivers("delete_policy_action_precommit", context)

    def delete_policy_action_postcommit(self, context):
        self._call_on_drivers("delete_policy_action_postcommit", context,
                              continue_on_failure=True)

    def get_policy_action_status(self, context):
        self._call_on_drivers("get_policy_action_status", context)

    def create_policy_rule_precommit(self, context):
        self._call_on_drivers("create_policy_rule_precommit", context)

    def create_policy_rule_postcommit(self, context):
        self._call_on_drivers("create_policy_rule_postcommit", context)

    def update_policy_rule_precommit(self, context):
        self._call_on_drivers("update_policy_rule_precommit", context)

    def update_policy_rule_postcommit(self, context):
        self._call_on_drivers("update_policy_rule_postcommit", context)

    def delete_policy_rule_precommit(self, context):
        self._call_on_drivers("delete_policy_rule_precommit", context)

    def delete_policy_rule_postcommit(self, context):
        self._call_on_drivers("delete_policy_rule_postcommit", context,
                              continue_on_failure=True)

    def get_policy_rule_status(self, context):
        self._call_on_drivers("get_policy_rule_status", context)

    def create_policy_rule_set_precommit(self, context):
        self._call_on_drivers("create_policy_rule_set_precommit", context)

    def create_policy_rule_set_postcommit(self, context):
        self._call_on_drivers("create_policy_rule_set_postcommit", context)

    def update_policy_rule_set_precommit(self, context):
        self._call_on_drivers("update_policy_rule_set_precommit", context)

    def update_policy_rule_set_postcommit(self, context):
        self._call_on_drivers("update_policy_rule_set_postcommit", context)

    def delete_policy_rule_set_precommit(self, context):
        self._call_on_drivers("delete_policy_rule_set_precommit", context)

    def delete_policy_rule_set_postcommit(self, context):
        self._call_on_drivers("delete_policy_rule_set_postcommit", context,
                              continue_on_failure=True)

    def get_policy_rule_set_status(self, context):
        self._call_on_drivers("get_policy_rule_set_status", context)

    def create_external_segment_precommit(self, context):
        self._call_on_drivers("create_external_segment_precommit",
                              context)

    def create_external_segment_postcommit(self, context):
        self._call_on_drivers("create_external_segment_postcommit",
                              context)

    def update_external_segment_precommit(self, context):
        self._call_on_drivers("update_external_segment_precommit",
                              context)

    def update_external_segment_postcommit(self, context):
        self._call_on_drivers("update_external_segment_postcommit",
                              context)

    def delete_external_segment_precommit(self, context):
        self._call_on_drivers("delete_external_segment_precommit",
                              context)

    def delete_external_segment_postcommit(self, context):
        self._call_on_drivers("delete_external_segment_postcommit",
                              context, continue_on_failure=True)

    def get_external_segment_status(self, context):
        self._call_on_drivers("get_external_segment_status", context)

    def create_external_policy_precommit(self, context):
        self._call_on_drivers("create_external_policy_precommit",
                              context)

    def create_external_policy_postcommit(self, context):
        self._call_on_drivers("create_external_policy_postcommit",
                              context)

    def update_external_policy_precommit(self, context):
        self._call_on_drivers("update_external_policy_precommit",
                              context)

    def update_external_policy_postcommit(self, context):
        self._call_on_drivers("update_external_policy_postcommit",
                              context)

    def delete_external_policy_precommit(self, context):
        self._call_on_drivers("delete_external_policy_precommit",
                              context)

    def delete_external_policy_postcommit(self, context):
        self._call_on_drivers("delete_external_policy_postcommit",
                              context, continue_on_failure=True)

    def get_external_policy_status(self, context):
        self._call_on_drivers("get_external_policy_status", context)

    def create_nat_pool_precommit(self, context):
        self._call_on_drivers("create_nat_pool_precommit", context)

    def create_nat_pool_postcommit(self, context):
        self._call_on_drivers("create_nat_pool_postcommit", context)

    def update_nat_pool_precommit(self, context):
        self._call_on_drivers("update_nat_pool_precommit", context)

    def update_nat_pool_postcommit(self, context):
        self._call_on_drivers("update_nat_pool_postcommit", context)

    def delete_nat_pool_precommit(self, context):
        self._call_on_drivers("delete_nat_pool_precommit", context)

    def delete_nat_pool_postcommit(self, context):
        self._call_on_drivers("delete_nat_pool_postcommit", context,
                              continue_on_failure=True)

    def get_nat_pool_status(self, context):
        self._call_on_drivers("get_nat_pool_status", context)

    def validate_state(self, repair):
        result = api.VALIDATION_PASSED
        for driver in self.ordered_policy_drivers:
            this_result = driver.obj.validate_state(repair)
            if (this_result == api.VALIDATION_FAILED_UNREPAIRABLE or
                (this_result == api.VALIDATION_FAILED_REPAIRABLE and
                 result != api.VALIDATION_FAILED_UNREPAIRABLE) or
                (this_result == api.VALIDATION_REPAIRED and
                 result == api.VALIDATION_PASSED)):
                result = this_result
        return result



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice._i18n import _LE
from gbpservice._i18n import _LI
from gbpservice.neutron.plugins.ml2plus import driver_api

from neutron.db import api as db_api
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import managers
from neutron.quota import resource_registry
from oslo_log import log
from oslo_utils import excutils

LOG = log.getLogger(__name__)


class MechanismManager(managers.MechanismManager):

    def __init__(self):
        super(MechanismManager, self).__init__()

    def _call_on_drivers(self, method_name, context,
            continue_on_failure=False, raise_db_retriable=False):
        super(MechanismManager, self)._call_on_drivers(
                method_name, context, continue_on_failure=False,
                raise_db_retriable=False)
        if method_name.endswith('_precommit'):
            # This does the same thing as:
            # https://github.com/openstack/neutron/blob/newton-eol/neutron/
            # api/v2/base.py#L489
            # but from within the scope of the plugin's transaction, such
            # that if it fails, everything that happened prior to this in
            # precommit phase can also be rolled back.
            resource_name = method_name.replace('_precommit', '').replace(
                    'create_', '').replace('update_', '').replace(
                            'delete_', '')
            tracked_resource = resource_registry.get_resource(resource_name)
            tracked_resource._dirty_tenants.add(context.current['tenant_id'])
            resource_registry.set_resources_dirty(context._plugin_context)

    def _call_on_extended_drivers(self, method_name, context,
                                  continue_on_failure=False,
                                  raise_db_retriable=False):
        """Call a method on all extended mechanism drivers.

        :param method_name: name of the method to call
        :param context: context parameter to pass to each method call
        :param continue_on_failure: whether or not to continue to call
        all mechanism drivers once one has raised an exception
        :param raise_db_retriable: whether or not to treat retriable db
        exception by mechanism drivers to propagate up to upper layer so
        that upper layer can handle it or error in ML2 player
        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver call fails. or DB retriable error when
        raise_db_retriable=False. See neutron.db.api.is_retriable for
        what db exception is retriable
        """
        errors = []
        for driver in self.ordered_mech_drivers:
            if isinstance(driver.obj, driver_api.MechanismDriver):
                try:
                    getattr(driver.obj, method_name)(context)
                except Exception as e:
                    if raise_db_retriable and db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug("DB exception raised by Mechanism "
                                      "driver '%(name)s' in %(method)s",
                                      {'name': driver.name,
                                       'method': method_name},
                                      exc_info=e)
                    LOG.exception(
                        _LE("Mechanism driver '%(name)s' failed in "
                            "%(method)s"),
                        {'name': driver.name, 'method': method_name}
                    )
                    errors.append(e)
                    if not continue_on_failure:
                        break
        if errors:
            raise ml2_exc.MechanismDriverError(
                method=method_name,
                errors=errors
            )

    def ensure_tenant(self, plugin_context, tenant_id):
        for driver in self.ordered_mech_drivers:
            if isinstance(driver.obj, driver_api.MechanismDriver):
                try:
                    driver.obj.ensure_tenant(plugin_context, tenant_id)
                except Exception as e:
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug("DB exception raised by Mechanism "
                                      "driver '%(name)s' in ensure_tenant",
                                      {'name': driver.name},
                                      exc_info=e)
                    LOG.exception(_LE("Mechanism driver '%s' failed in "
                                      "ensure_tenant"), driver.name)
                    raise ml2_exc.MechanismDriverError(method="ensure_tenant")

    def create_subnetpool_precommit(self, context):
        self._call_on_extended_drivers("create_subnetpool_precommit",
                                       context, raise_db_retriable=True)

    def create_subnetpool_postcommit(self, context):
        self._call_on_extended_drivers("create_subnetpool_postcommit",
                                       context)

    def update_subnetpool_precommit(self, context):
        self._call_on_extended_drivers("update_subnetpool_precommit",
                                       context, raise_db_retriable=True)

    def update_subnetpool_postcommit(self, context):
        self._call_on_extended_drivers("update_subnetpool_postcommit",
                                       context, continue_on_failure=True)

    def delete_subnetpool_precommit(self, context):
        self._call_on_extended_drivers("delete_subnetpool_precommit",
                                       context, raise_db_retriable=True)

    def delete_subnetpool_postcommit(self, context):
        self._call_on_extended_drivers("delete_subnetpool_postcommit",
                                       context, continue_on_failure=True)

    def create_address_scope_precommit(self, context):
        self._call_on_extended_drivers("create_address_scope_precommit",
                                       context, raise_db_retriable=True)

    def create_address_scope_postcommit(self, context):
        self._call_on_extended_drivers("create_address_scope_postcommit",
                                       context)

    def update_address_scope_precommit(self, context):
        self._call_on_extended_drivers("update_address_scope_precommit",
                                       context, raise_db_retriable=True)

    def update_address_scope_postcommit(self, context):
        self._call_on_extended_drivers("update_address_scope_postcommit",
                                       context, continue_on_failure=True)

    def delete_address_scope_precommit(self, context):
        self._call_on_extended_drivers("delete_address_scope_precommit",
                                       context, raise_db_retriable=True)

    def delete_address_scope_postcommit(self, context):
        self._call_on_extended_drivers("delete_address_scope_postcommit",
                                       context, continue_on_failure=True)

    def create_security_group_precommit(self, context):
        self._call_on_extended_drivers("create_security_group_precommit",
                                       context, raise_db_retriable=True)

    def create_security_group_postcommit(self, context):
        self._call_on_extended_drivers("create_security_group_postcommit",
                                       context)

    def update_security_group_precommit(self, context):
        self._call_on_extended_drivers("update_security_group_precommit",
                                       context, raise_db_retriable=True)

    def update_security_group_postcommit(self, context):
        self._call_on_extended_drivers("update_security_group_postcommit",
                                       context, continue_on_failure=True)

    def delete_security_group_precommit(self, context):
        self._call_on_extended_drivers("delete_security_group_precommit",
                                       context, raise_db_retriable=True)

    def delete_security_group_postcommit(self, context):
        self._call_on_extended_drivers("delete_security_group_postcommit",
                                       context, continue_on_failure=True)

    def create_security_group_rule_precommit(self, context):
        self._call_on_extended_drivers("create_security_group_rule_precommit",
                                       context, raise_db_retriable=True)

    def create_security_group_rule_postcommit(self, context):
        self._call_on_extended_drivers("create_security_group_rule_postcommit",
                                       context)

    def update_security_group_rule_precommit(self, context):
        self._call_on_extended_drivers("update_security_group_rule_precommit",
                                       context, raise_db_retriable=True)

    def update_security_group_rule_postcommit(self, context):
        self._call_on_extended_drivers("update_security_group_rule_postcommit",
                                       context, continue_on_failure=True)

    def delete_security_group_rule_precommit(self, context):
        self._call_on_extended_drivers("delete_security_group_rule_precommit",
                                       context, raise_db_retriable=True)

    def delete_security_group_rule_postcommit(self, context):
        self._call_on_extended_drivers("delete_security_group_rule_postcommit",
                                       context, continue_on_failure=True)


class ExtensionManager(managers.ExtensionManager):

    def __init__(self):
        super(ExtensionManager, self).__init__()

    def _call_on_extended_drivers(self, method_name, plugin_context, data,
                                  result):
        """Call a method on all extended extension drivers."""
        for driver in self.ordered_ext_drivers:
            if isinstance(driver.obj, driver_api.ExtensionDriver):
                try:
                    getattr(driver.obj, method_name)(plugin_context, data,
                                                     result)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        LOG.info(_LI("Extension driver '%(name)s' failed in "
                                     "%(method)s"),
                                 {'name': driver.name, 'method': method_name})

    # Overrides ML2 implementation to avoid eating retriable
    # exceptions, as well as to support calling only on extension
    # drivers extended for ML2Plus.
    def _call_on_dict_driver(self, method_name, session, base_model, result,
                             extended_only=False, has_base_model=True):

        # Bulk operations might not be implemented by all drivers
        def noop(*args, **kwargs):
            pass

        for driver in self.ordered_ext_drivers:
            if not extended_only or isinstance(
                    driver.obj, driver_api.ExtensionDriver):
                try:
                    if not has_base_model:
                        getattr(driver.obj, method_name, noop)(session,
                                                               result)
                    else:
                        getattr(driver.obj, method_name, noop)(session,
                                                               base_model,
                                                               result)
                except Exception as e:
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug(
                                "DB exception raised by extension driver "
                                "'%(name)s' in %(method)s",
                                {'name': driver.name, 'method': method_name},
                                exc_info=e)
                    LOG.exception(
                        "Extension driver '%(name)s' failed in %(method)s",
                        {'name': driver.name, 'method': method_name})
                    raise ml2_exc.ExtensionDriverError(driver=driver.name)

    def process_create_subnetpool(self, plugin_context, data, result):
        self._call_on_extended_drivers("process_create_subnetpool",
                                       plugin_context, data, result)

    def process_update_subnetpool(self, plugin_context, data, result):
        self._call_on_extended_drivers("process_update_subnetpool",
                                       plugin_context, data, result)

    def extend_subnetpool_dict(self, session, base_model, result):
        self._call_on_dict_driver("extend_subnetpool_dict",
                                  session, base_model, result, True)

    def process_create_address_scope(self, plugin_context, data, result):
        self._call_on_extended_drivers("process_create_address_scope",
                                       plugin_context, data, result)

    def process_update_address_scope(self, plugin_context, data, result):
        self._call_on_extended_drivers("process_update_address_scope",
                                       plugin_context, data, result)

    def extend_address_scope_dict(self, session, base_model, result):
        self._call_on_dict_driver("extend_address_scope_dict",
                                  session, base_model, result, True)

    def extend_subnetpool_dict_bulk(self, session, result):
        self._call_on_dict_driver("extend_subnetpool_dict_bulk",
                                  session, None, result, True,
                                  has_base_model=False)

    def extend_address_scope_dict_bulk(self, session, result):
        self._call_on_dict_driver("extend_address_scope_dict_bulk",
                                  session, None, result, True,
                                  has_base_model=False)

    def extend_network_dict_bulk(self, session, result):
        self._call_on_dict_driver("extend_network_dict_bulk", session,
                                  None, result, has_base_model=False)

    def extend_subnet_dict_bulk(self, session, result):
        self._call_on_dict_driver("extend_subnet_dict_bulk", session,
                                  None, result, has_base_model=False)

    def extend_port_dict_bulk(self, session, result):
        self._call_on_dict_driver("extend_port_dict_bulk", session, None,
                                  result, has_base_model=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

_uuid = uuidutils.generate_uuid


def gbp_attributes(func):
    def inner(**kwargs):
        attrs = func()
        attrs.update(kwargs)
        project_id = _uuid()
        if 'prj' in func.__name__ or 'default' not in func.__name__ and (
            'update' not in func.__name__):
            attrs.update({'project_id': project_id, 'tenant_id': project_id})
        return attrs
    return inner


@gbp_attributes
def get_create_policy_target_default_attrs():
    return {'name': '', 'description': '', 'policy_target_group_id': None,
            'cluster_id': ''}


@gbp_attributes
def get_create_policy_target_attrs():
    return {'name': 'ep1', 'policy_target_group_id': _uuid(),
            'description': 'test policy_target',
            'cluster_id': 'some_cluster_id'}


@gbp_attributes
def get_update_policy_target_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_application_policy_group_default_attrs():
    return {'name': '', 'description': '', 'shared': False}


@gbp_attributes
def get_create_application_policy_group_attrs():
    return {'name': 'apg1', 'tenant_id': _uuid(),
            'description': 'test application_policy_group',
            'shared': False}


@gbp_attributes
def get_update_application_policy_group_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_policy_target_group_default_attrs():
    return {'name': '', 'description': '', 'l2_policy_id': None,
            'application_policy_group_id': None,
            'provided_policy_rule_sets': {},
            'consumed_policy_rule_sets': {},
            'network_service_policy_id': None, 'shared': False,
            'service_management': False}


@gbp_attributes
def get_create_policy_target_group_attrs():
    return {'name': 'ptg1',
            'description': 'test policy_target group',
            'l2_policy_id': _uuid(),
            'application_policy_group_id': _uuid(),
            'provided_policy_rule_sets': {_uuid(): None},
            'consumed_policy_rule_sets': {_uuid(): None},
            'network_service_policy_id': _uuid(),
            'shared': False, 'service_management': False}


@gbp_attributes
def get_update_policy_target_group_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_l2_policy_default_attrs():
    return {'name': '', 'description': '', 'shared': False,
            'inject_default_route': True}


@gbp_attributes
def get_create_l2_policy_attrs():
    return {'name': 'l2p1',
            'description': 'test L2 policy', 'l3_policy_id': _uuid(),
            'inject_default_route': True, 'shared': False}


@gbp_attributes
def get_update_l2_policy_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_l3_policy_default_attrs():
    return {'name': '', 'description': '', 'ip_version': 4,
            'ip_pool': '10.0.0.0/8', 'subnet_prefix_length': 24,
            'external_segments': {}, 'shared': False}


@gbp_attributes
def get_create_l3_policy_attrs():
    return {'name': 'l3p1',
            'description': 'test L3 policy', 'ip_version': 6,
            'ip_pool': 'fd01:2345:6789::/48',
            'external_segments': {_uuid(): ['192.168.0.3']},
            'subnet_prefix_length': 64, 'shared': False}


@gbp_attributes
def get_update_l3_policy_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_policy_action_default_attrs():
    return {'name': '',
            'description': '',
            'action_type': 'allow',
            'action_value': None,
            'shared': False}


@gbp_attributes
def get_create_policy_action_attrs():
    return {'name': 'pa1',
            'description': 'test policy action',
            'action_type': 'redirect',
            'action_value': _uuid(),
            'shared': False}


@gbp_attributes
def get_update_policy_action_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_policy_classifier_default_attrs():
    return {'name': '',
            'description': '',
            'protocol': None,
            'port_range': None,
            'direction': None,
            'shared': False}


@gbp_attributes
def get_create_policy_classifier_attrs():
    return {'name': 'pc1',
            'description': 'test policy classifier',
            'protocol': 'tcp',
            'port_range': '100:200',
            'direction': 'in',
            'shared': False}


@gbp_attributes
def get_update_policy_classifier_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_policy_rule_default_attrs():
    return {'name': '',
            'description': '',
            'enabled': True,
            'policy_actions': [],
            'shared': False}


@gbp_attributes
def get_create_policy_rule_attrs():
    return {'name': 'pr1',
            'description': 'test policy rule',
            'enabled': True,
            'policy_classifier_id': _uuid(),
            'policy_actions': [_uuid()],
            'shared': False}


@gbp_attributes
def get_update_policy_rule_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_policy_rule_set_default_attrs():
    return {'name': '',
            'description': '',
            'child_policy_rule_sets': [],
            'policy_rules': [],
            'shared': False}


@gbp_attributes
def get_create_policy_rule_set_attrs():
    return {'name': 'policy_rule_set1',
            'description': 'test policy_rule_set',
            'child_policy_rule_sets': [_uuid()],
            'policy_rules': [_uuid()],
            'shared': False}


@gbp_attributes
def get_update_policy_rule_set_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_network_service_policy_default_attrs():
    return {'name': '', 'description': '',
            'network_service_params': [], 'shared': False}


@gbp_attributes
def get_create_network_service_policy_attrs():
    return {'name': 'nsp1',
            'shared': False,
            'description': 'test Net Svc Policy',
            'network_service_params': [{'type': 'ip_single', 'name': 'vip',
                                        'value': 'self_subnet'}]}


@gbp_attributes
def get_update_network_service_policy_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_external_policy_default_attrs():
    return {'name': '', 'description': '',
            'external_segments': [],
            'provided_policy_rule_sets': {},
            'consumed_policy_rule_sets': {},
            'shared': False}


@gbp_attributes
def get_create_external_policy_attrs():
    return {'name': 'ep1',
            'description': 'test ep',
            'external_segments': [_uuid()],
            'provided_policy_rule_sets': {_uuid(): None},
            'consumed_policy_rule_sets': {_uuid(): None},
            'shared': False}


@gbp_attributes
def get_update_external_policy_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_external_segment_default_attrs():
    return {'name': '', 'description': '',
            'external_routes': [],
            'ip_version': 4,
            'cidr': '172.16.0.0/12',
            'port_address_translation': False,
            'shared': False}


@gbp_attributes
def get_create_external_segment_attrs():
    return {'name': 'es1',
            'description': 'test ep',
            'external_routes': [{'destination': '0.0.0.0/0',
                                 'nexthop': '192.168.0.1'}],
            'cidr': '192.168.0.0/24',
            'ip_version': 4, 'port_address_translation': True,
            'shared': False}


@gbp_attributes
def get_update_external_segment_attrs():
    return {'name': 'new_name'}


@gbp_attributes
def get_create_nat_pool_default_attrs():
    return {'name': '', 'description': '',
            'external_segment_id': None, 'ip_version': 4,
            'ip_pool': '172.16.0.0/16',
            'shared': False}


@gbp_attributes
def get_create_nat_pool_attrs():
    return {'name': 'es1',
            'description': 'test ep',
            'ip_version': 4,
            'ip_pool': '172.16.0.0/16',
            'external_segment_id': _uuid(),
            'shared': False}


@gbp_attributes
def get_update_nat_pool_attrs():
    return {'name': 'new_name'}


# Service Chain
@gbp_attributes
def get_create_service_profile_default_attrs():
    return {'name': '', 'description': ''}


@gbp_attributes
def get_create_service_profile_attrs():
    return {
        'name': 'serviceprofile1',
        'service_type': 'FIREWALL',
        'description': 'test service profile',
    }


@gbp_attributes
def get_update_service_profile_attrs():
    return {
        'name': 'new_name',
    }


@gbp_attributes
def get_create_servicechain_node_default_attrs():
    return {
        'name': '',
        'description': '',
        'config': '{}',
        'service_type': None,
        'shared': False,
    }


@gbp_attributes
def get_create_servicechain_node_attrs():
    return {
        'name': 'servicechain1',
        'service_profile_id': _uuid(),
        'description': 'test servicechain node',
        'config': '{}',
        'service_type': None,
        'shared': True,
    }


@gbp_attributes
def get_update_servicechain_node_attrs():
        return {
            'name': 'new_name',
            'config': 'new_config',
        }


@gbp_attributes
def get_create_servicechain_spec_default_attrs():
    return {
        'name': '',
        'description': '',
        'nodes': [],
        'shared': False,
    }


@gbp_attributes
def get_create_servicechain_spec_attrs():
    return {
        'name': 'servicechainspec1',
        'nodes': [_uuid(), _uuid()],
        'description': 'test servicechain spec',
        'shared': True,
    }


@gbp_attributes
def get_update_servicechain_spec_attrs():
    return {
        'name': 'new_name',
        'nodes': [_uuid()]
    }


@gbp_attributes
def get_create_servicechain_instance_default_attrs():
    return {'name': '', 'description': '', 'config_param_values': "{}"}


@gbp_attributes
def get_create_servicechain_instance_attrs():
    return {
        'name': 'servicechaininstance1',
        'servicechain_specs': [_uuid()],
        'provider_ptg_id': _uuid(),
        'consumer_ptg_id': _uuid(),
        'management_ptg_id': _uuid(),
        'classifier_id': _uuid(),
        'config_param_values': "{}",
        'description': 'test servicechain instance'
    }


def get_update_servicechain_instance_attrs():
    return {
        'name': 'new_name',
        'servicechain_specs': [_uuid()],
        'classifier_id': _uuid()
    }


@gbp_attributes
def get_create_application_policy_group_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'shared': False}


@gbp_attributes
def get_create_policy_target_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'policy_target_group_id': None,
            'cluster_id': ''}


@gbp_attributes
def get_create_policy_target_group_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'l2_policy_id': None,
            'application_policy_group_id': None,
            'provided_policy_rule_sets': {},
            'consumed_policy_rule_sets': {},
            'network_service_policy_id': None, 'shared': False,
            'service_management': False}


@gbp_attributes
def get_create_l2_policy_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'shared': False,
            'inject_default_route': True}


@gbp_attributes
def get_create_l3_policy_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'ip_version': 4,
            'ip_pool': '10.0.0.0/8', 'subnet_prefix_length': 24,
            'external_segments': {}, 'shared': False}


@gbp_attributes
def get_create_policy_action_default_attrs_and_prj_id():
    return {'name': '',
            'description': '',
            'action_type': 'allow',
            'action_value': None,
            'shared': False}


@gbp_attributes
def get_create_policy_classifier_default_attrs_and_prj_id():
    return {'name': '',
            'description': '',
            'protocol': None,
            'port_range': None,
            'direction': None,
            'shared': False}


@gbp_attributes
def get_create_policy_rule_default_attrs_and_prj_id():
    return {'name': '',
            'description': '',
            'enabled': True,
            'policy_actions': [],
            'shared': False}


@gbp_attributes
def get_create_policy_rule_set_default_attrs_and_prj_id():
    return {'name': '',
            'description': '',
            'child_policy_rule_sets': [],
            'policy_rules': [],
            'shared': False}


@gbp_attributes
def get_create_network_service_policy_default_attrs_and_prj_id():
    return {'name': '', 'description': '',
            'network_service_params': [], 'shared': False}


@gbp_attributes
def get_create_external_policy_default_attrs_and_prj_id():
    return {'name': '', 'description': '',
            'external_segments': [],
            'provided_policy_rule_sets': {},
            'consumed_policy_rule_sets': {},
            'shared': False}


@gbp_attributes
def get_create_external_segment_default_attrs_and_prj_id():
    return {'name': '', 'description': '',
            'external_routes': [],
            'ip_version': 4,
            'cidr': '172.16.0.0/12',
            'port_address_translation': False,
            'shared': False}


@gbp_attributes
def get_create_nat_pool_default_attrs_and_prj_id():
    return {'name': '', 'description': '',
            'external_segment_id': None, 'ip_version': 4,
            'ip_pool': '172.16.0.0/16',
            'shared': False}


# Service Chain
@gbp_attributes
def get_create_service_profile_default_attrs_and_prj_id():
    return {'name': '', 'description': ''}


@gbp_attributes
def get_create_servicechain_node_default_attrs_and_prj_id():
    return {
        'name': '',
        'description': '',
        'config': '{}',
        'service_type': None,
        'shared': False,
    }


@gbp_attributes
def get_create_servicechain_spec_default_attrs_and_prj_id():
    return {
        'name': '',
        'description': '',
        'nodes': [],
        'shared': False,
    }


@gbp_attributes
def get_create_servicechain_instance_default_attrs_and_prj_id():
    return {'name': '', 'description': '', 'config_param_values': "{}"}


def get_resource_plural(resource):
    if resource.endswith('y'):
        resource_plural = resource.replace('y', 'ies')
    else:
        resource_plural = resource + 's'

    return resource_plural



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Group Policy Resource Mapping Driver (gpm_rmd_1)

Revision ID: 1bf7555fa01a
Revises: 23b6c4d703c7
Create Date: 2014-07-24 16:12:22.610815

"""

# revision identifiers, used by Alembic.
revision = '1bf7555fa01a'
down_revision = '23b6c4d703c7'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gpm_owned_networks',
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id')
    )

    op.create_table(
        'gpm_owned_ports',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id')
    )

    op.create_table(
        'gpm_owned_subnets',
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('subnet_id')
    )

    op.create_table(
        'gpm_owned_routers',
        sa.Column('router_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('router_id')
    )


def downgrade():

    op.drop_table('gpm_owned_routers')
    op.drop_table('gpm_owned_subnets')
    op.drop_table('gpm_owned_ports')
    op.drop_table('gpm_owned_networks')



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions

from gbpservice.neutron.plugins.ml2plus import extension_overrides


# Monkeypatch Neutron to allow overriding its own extension
# descriptors. Note that extension descriptor classes cannot be
# monkeypatched directly because they are loaded explicitly by file
# name and then used immediately.
_real_get_extensions_path = extensions.get_extensions_path


def get_extensions_path(service_plugins=None):
    path = _real_get_extensions_path(service_plugins)
    return extension_overrides.__path__[0] + ':' + path


extensions.get_extensions_path = get_extensions_path


import sys

from neutron import context as n_context

from gbpservice.network.neutronv2 import local_api


def get_current_session():
    i = 1
    not_found = True
    try:
        while not_found:
            for val in sys._getframe(i).f_locals.itervalues():
                if isinstance(val, n_context.Context):
                    ctx = val
                    not_found = False
                    break
            i = i + 1
        return ctx.session
    except Exception:
        return


from neutron.callbacks import registry


def notify(resource, event, trigger, **kwargs):
    if 'context' in kwargs:
        session = kwargs['context'].session
    else:
        session = get_current_session()

    txn = None
    if session:
        txn = local_api.get_outer_transaction(session.transaction)
    local_api.send_or_queue_registry_notification(
        session, txn, resource, event, trigger, **kwargs)


registry.notify = notify


from neutron._i18n import _LE
from neutron.callbacks import events
from neutron.callbacks import exceptions
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def _notify_loop(resource, event, trigger, **kwargs):
    """The notification loop."""
    errors = []
    callbacks = kwargs.pop('callbacks', None)
    if not callbacks:
        callbacks = list(registry._get_callback_manager()._callbacks[
            resource].get(event, {}).items())
    LOG.debug("Notify callbacks %s for %s, %s", callbacks, resource, event)
    for callback_id, callback in callbacks:
        try:
            callback(resource, event, trigger, **kwargs)
        except Exception as e:
            abortable_event = (
                event.startswith(events.BEFORE) or
                event.startswith(events.PRECOMMIT)
            )
            if not abortable_event:
                LOG.exception(_LE("Error during notification for "
                                  "%(callback)s %(resource)s, %(event)s"),
                              {'callback': callback_id,
                               'resource': resource, 'event': event})
            else:
                LOG.error(_LE("Callback %(callback)s raised %(error)s"),
                          {'callback': callback_id, 'error': e})
            errors.append(exceptions.NotificationError(callback_id, e))
    return errors


original_notify_loop = registry._get_callback_manager()._notify_loop


from inspect import isclass
from inspect import isfunction
from inspect import ismethod


# The undecorated() and looks_like_a_decorator() functions have been
# borrowed from the undecorated python library since RPM or Debian
# packages are not readily available.
def looks_like_a_decorator(a):
    return (
        isfunction(a) or ismethod(a) or isclass(a)
    )


def undecorated(o):
    """Remove all decorators from a function, method or class"""
    # class decorator
    if type(o) is type:
        return o

    try:
        # python2
        closure = o.func_closure
    except AttributeError:
        pass

    try:
        # python3
        closure = o.__closure__
    except AttributeError:
        return

    if closure:
        for cell in closure:
            # avoid infinite recursion
            if cell.cell_contents is o:
                continue

            # check if the contents looks like a decorator; in that case
            # we need to go one level down into the dream, otherwise it
            # might just be a different closed-over variable, which we
            # can ignore.

            # Note: this favors supporting decorators defined without
            # @wraps to the detriment of function/method/class closures
            if looks_like_a_decorator(cell.cell_contents):
                undecd = undecorated(cell.cell_contents)
                if undecd:
                    return undecd
        else:
            return o
    else:
        return o

from neutron.db import common_db_mixin as common_db_api
from neutron.db.quota import api as quota_api
from neutron.db.quota import driver  # noqa
from neutron.db.quota import models as quota_models
from neutron import quota
from neutron.quota import resource_registry as res_reg
from oslo_config import cfg


f = quota_api.remove_reservation
quota_api.commit_reservation = undecorated(f)


def commit_reservation(context, reservation_id):
    quota_api.commit_reservation(context, reservation_id, set_dirty=False)


quota.QUOTAS.get_driver().commit_reservation = commit_reservation


def patched_set_resources_dirty(context):
    if not cfg.CONF.QUOTAS.track_quota_usage:
        return

    with context.session.begin(subtransactions=True):
        for res in res_reg.get_all_resources().values():
            if res_reg.is_tracked(res.name) and res.dirty:
                dirty_tenants_snap = res._dirty_tenants.copy()
                for tenant_id in dirty_tenants_snap:
                    query = common_db_api.model_query(
                            context, quota_models.QuotaUsage)
                    query = query.filter_by(resource=res.name).filter_by(
                            tenant_id=tenant_id)
                    usage_data = query.first()
                    # Set dirty if not set already. This effectively
                    # patches the inner notify method:
                    # https://github.com/openstack/neutron/blob/newton-eol/
                    # neutron/api/v2/base.py#L481
                    # to avoid updating the QuotaUsages table outside
                    # from that method (which starts a new transaction).
                    # The dirty marking would have been already done
                    # in the ml2plus manager at the end of the pre_commit
                    # stage (and prior to the plugin initiated transaction
                    # completing).
                    if usage_data and not usage_data.dirty:
                        res.mark_dirty(context)


quota.resource_registry.set_resources_dirty = patched_set_resources_dirty


from neutron._i18n import _LI
from oslo_db.sqlalchemy import exc_filters


exc_filters._LE = _LI
exc_filters.LOG.exception = exc_filters.LOG.debug


from neutron.db import models_v2
from neutron.plugins.ml2 import db as ml2_db
from neutron.plugins.ml2 import models
from sqlalchemy.orm import exc


# REVISIT: This method gets decorated in Pike for removal in Queens. So this
# patching might need to be changed in Pike and removed in Queens.
def patched_get_locked_port_and_binding(session, port_id):
    """Get port and port binding records for update within transaction."""
    LOG.debug("Using patched_get_locked_port_and_binding")
    try:
        port = (session.query(models_v2.Port).
                enable_eagerloads(False).
                filter_by(id=port_id).
                one())
        binding = (session.query(models.PortBinding).
                   enable_eagerloads(False).
                   filter_by(port_id=port_id).
                   one())
        return port, binding
    except exc.NoResultFound:
        return None, None


ml2_db.get_locked_port_and_binding = patched_get_locked_port_and_binding


from neutron.db import db_base_plugin_v2


DEVICE_OWNER_SVI_PORT = 'apic:svi'
db_base_plugin_v2.AUTO_DELETE_PORT_OWNERS.append(DEVICE_OWNER_SVI_PORT)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import subprocess

import netaddr
import netifaces
from oslo_log._i18n import _LE
from oslo_log._i18n import _LI
from oslo_log import log as logging
import oslo_serialization.jsonutils as jsonutils
import pecan
from pecan import rest
import time
import yaml

LOG = logging.getLogger(__name__)
SUCCESS = 'SUCCESS'
FAILED = 'FAILED'

notifications = []
FW_SCRIPT_PATH = ("/usr/local/lib/python2.7/dist-packages/" +
                  "gbpservice/tests/contrib/nfp_service/" +
                  "reference_configurator/scripts/configure_fw_rules.py")


class Controller(rest.RestController):

    """Implements all the APIs Invoked by HTTP requests.

    Implements following HTTP methods.
        -get
        -post

    """
    def __init__(self, method_name):
        try:
            self.method_name = "network_function_device_notification"
            super(Controller, self).__init__()
            ip_a = subprocess.Popen('ifconfig -a', shell=True,
                                    stdout=subprocess.PIPE).stdout.read()
            out1 = subprocess.Popen('dhclient eth0', shell=True,
                                    stdout=subprocess.PIPE).stdout.read()
            out2 = subprocess.Popen('dhclient eth0', shell=True,
                                    stdout=subprocess.PIPE).stdout.read()
            output = "%s\n%s\n%s" % (ip_a, out1, out2)
            LOG.info(_LI("Dhclient on eth0, result: %(output)s"),
                     {'output': output})
        except Exception as err:
            msg = (
                "Failed to initialize Controller class  %s." %
                str(err).capitalize())
            LOG.error(msg)

    def _push_notification(self, context,
                           notification_data, service_type):
        response = {'info': {'service_type': service_type,
                             'context': context},
                    'notification': notification_data
                    }

        notifications.append(response)

    @pecan.expose(method='GET', content_type='application/json')
    def get(self):
        """Method of REST server to handle request get_notifications.

        This method send an RPC call to configurator and returns Notification
        data to config-agent

        Returns: Dictionary that contains Notification data

        """

        global notifications
        try:
            notification_data = jsonutils.dumps(notifications)
            msg = ("NOTIFICATION_DATA sent to config_agent %s"
                   % notification_data)
            LOG.info(msg)
            notifications = []
            return notification_data
        except Exception as err:
            pecan.response.status = 500
            msg = ("Failed to get notification_data  %s."
                   % str(err).capitalize())
            LOG.error(msg)
            error_data = self._format_description(msg)
            return jsonutils.dumps(error_data)

    @pecan.expose(method='POST', content_type='application/json')
    def post(self, **body):
        try:
            body = None
            if pecan.request.is_body_readable:
                body = pecan.request.json_body

            msg = ("Request data:: %s" % body)
            LOG.debug(msg)

            config_datas = body['config']
            service_type = body['info']['service_type']
            notification_data = []

            for config_data in config_datas:
                try:
                    resource = config_data['resource']
                    if resource == 'healthmonitor':
                        self._configure_healthmonitor(config_data)
                    elif resource == 'interfaces':
                        self._configure_interfaces(config_data)
                    elif resource == 'routes':
                        self._add_routes(config_data)
                    elif (config_data['resource'] in ['ansible', 'heat',
                                                      'custom_json']):
                        self._apply_user_config(config_data)
                    else:
                        status_msg = 'Unsupported resource'
                        notification_data.append(
                                {'resource': resource,
                                 'data': {'status_code': FAILED,
                                          'status_msg': status_msg}})
                    notification_data.append(
                                {'resource': config_data['resource'],
                                 'data': {'status_code': SUCCESS}})
                except Exception as ex:
                    notification_data.append(
                                {'resource': resource,
                                 'data': {'status_code': FAILED,
                                          'status_msg': str(ex)}})

            context = body['info']['context']
            self._push_notification(context, notification_data,
                                    service_type)
        except Exception as err:
            pecan.response.status = 500
            msg = ("Failed to serve HTTP post request %s %s."
                   % (self.method_name, str(err).capitalize()))
            LOG.error(msg)
            error_data = self._format_description(msg)
            return jsonutils.dumps(error_data)

    def _format_description(self, msg):
        """This method formats error description.

        :param msg: An error message that is to be formatted

        Returns: error_data dictionary
        """

        return {'failure_desc': {'msg': msg}}

    def _configure_healthmonitor(self, config_data):
        LOG.info(_LI("Configures healthmonitor with configuration "
                 "data : %(healthmonitor_data)s "),
                 {'healthmonitor_data': config_data})

    def _configure_interfaces(self, config_data):
        out1 = subprocess.Popen('sudo dhclient eth1', shell=True,
                                stdout=subprocess.PIPE).stdout.read()
        out2 = subprocess.Popen('sudo dhclient eth2', shell=True,
                                stdout=subprocess.PIPE).stdout.read()
        out3 = subprocess.Popen('cat /etc/network/interfaces', shell=True,
                                stdout=subprocess.PIPE).stdout.read()
        output = "%s\n%s\n%s" % (out1, out2, out3)
        LOG.info(_LI("Dhclient on eth0, result: %(initial_data)s"),
                 {'initial_data': output})
        LOG.info(_LI("Configures interfaces with configuration "
                 "data : %(interface_data)s "),
                 {'interface_data': config_data})

    def get_source_cidrs_and_gateway_ip(self, route_info):
        nfds = route_info['resource_data']['nfds']
        source_cidrs = []
        for nfd in nfds:
            for network in nfd['networks']:
                source_cidrs.append(network['cidr'])
                if network['type'] == 'stitching':
                    gateway_ip = network['gw_ip']
        return source_cidrs, gateway_ip

    def _add_routes(self, route_info):
        LOG.info(_LI("Configuring routes with configuration "
                 "data : %(route_data)s "),
                 {'route_data': route_info['resource_data']})
        source_cidrs, gateway_ip = self.get_source_cidrs_and_gateway_ip(
                                        route_info)
        default_route_commands = []
        for cidr in source_cidrs:
            try:
                source_interface = self._get_if_name_by_cidr(cidr)
            except Exception:
                raise Exception("Some of the interfaces do not have "
                                "IP Address")
            try:
                interface_number_string = source_interface.split("eth", 1)[1]
            except IndexError:
                LOG.error(_LE("Retrieved wrong interface %(interface)s for "
                          "configuring routes"),
                          {'interface': source_interface})
            try:
                routing_table_number = 20 + int(interface_number_string)

                ip_rule_command = "ip rule add from %s table %s" % (
                    cidr, routing_table_number)
                out1 = subprocess.Popen(ip_rule_command, shell=True,
                                        stdout=subprocess.PIPE).stdout.read()
                ip_rule_command = "ip rule add to %s table main" % (cidr)
                out2 = subprocess.Popen(ip_rule_command, shell=True,
                                        stdout=subprocess.PIPE).stdout.read()
                ip_route_command = "ip route add table %s default via %s" % (
                                        routing_table_number, gateway_ip)
                default_route_commands.append(ip_route_command)
                output = "%s\n%s" % (out1, out2)
                LOG.info(_LI("Static route configuration result: %(output)s"),
                         {'output': output})
            except Exception as ex:
                raise Exception("Failed to add static routes: %(ex)s" % {
                                'ex': str(ex)})
        for command in default_route_commands:
            try:
                out = subprocess.Popen(command, shell=True,
                                       stdout=subprocess.PIPE).stdout.read()
                LOG.info(_LI("Static route configuration result: %(output)s"),
                         {'output': out})
            except Exception as ex:
                raise Exception("Failed to add static routes: %(ex)s" % {
                                'ex': str(ex)})

    def _get_if_name_by_cidr(self, cidr):
        interfaces = netifaces.interfaces()
        retry_count = 0
        while True:
            all_interfaces_have_ip = True
            for interface in interfaces:
                inet_list = netifaces.ifaddresses(interface).get(
                    netifaces.AF_INET)
                if not inet_list:
                    all_interfaces_have_ip = False
                for inet_info in inet_list or []:
                    netmask = inet_info.get('netmask')
                    ip_address = inet_info.get('addr')
                    subnet_prefix = cidr.split("/")
                    if (ip_address == subnet_prefix[0] and (
                         len(subnet_prefix) == 1 or subnet_prefix[1] == "32")):
                        return interface
                    ip_address_netmask = '%s/%s' % (ip_address, netmask)
                    interface_cidr = netaddr.IPNetwork(ip_address_netmask)
                    if str(interface_cidr.cidr) == cidr:
                        return interface
            # Sometimes the hotplugged interface takes time to get IP
            if not all_interfaces_have_ip:
                if retry_count < 10:
                    time.sleep(3)
                    retry_count = retry_count + 1
                    continue
                else:
                    raise Exception("Some of the interfaces do not have "
                                    "IP Address")

    def _apply_user_config(self, config_data):
        LOG.info(_LI("Applying user config with configuration "
                 "type : %(config_type)s and "
                 "configuration data : %(config_data)s "),
                 {'config_type': config_data['resource'],
                  'config_data': config_data['resource_data']})
        service_config = config_data['resource_data'][
                                     'config_string']
        service_config = str(service_config)
        if config_data['resource'] == 'ansible':
            config_str = service_config.lstrip('ansible:')
            rules = config_str
        elif config_data['resource'] == 'heat':
            config_str = service_config.lstrip('heat_config:')
            rules = self._get_rules_from_config(config_str)
        elif config_data['resource'] == 'custom_json':
            config_str = service_config.lstrip('custom_json:')
            rules = config_str

        fw_rule_file = FW_SCRIPT_PATH
        command = ("sudo python " + fw_rule_file + " '" +
                   rules + "'")
        subprocess.check_output(command, stderr=subprocess.STDOUT,
                                shell=True)

    def _get_rules_from_config(self, config_str):
        rules_list = []
        try:
            stack_template = (jsonutils.loads(config_str) if
                              config_str.startswith('{') else
                              yaml.load(config_str))
        except Exception:
            return config_str

        resources = stack_template['resources']
        for resource in resources:
            if resources[resource]['type'] == 'OS::Neutron::FirewallRule':
                rule_info = {}
                destination_port = ''
                rule = resources[resource]['properties']
                protocol = rule['protocol']
                rule_info['action'] = 'log'
                rule_info['name'] = protocol
                if rule.get('destination_port'):
                    destination_port = rule['destination_port']
                if protocol == 'tcp':
                    rule_info['service'] = (protocol + '/' +
                                            str(destination_port))
                else:
                    rule_info['service'] = protocol
                rules_list.append(rule_info)

        return jsonutils.dumps({'rules': rules_list})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests
import subprocess

from oslo_serialization import jsonutils

from gbpservice.contrib.nfp.configurator.lib import constants as const
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


def set_class_attr(**kwargs):
    def f(class_obj):
        for key, value in kwargs.items():
            setattr(class_obj, key.lower(), value.lower())
        return class_obj
    return f


class BaseDriver(object):
    """ Implements common functions for drivers.

    Every service vendor must inherit this class. If any service vendor wants
    to add extra methods for their service, apart from below given, they should
    add method definition here and implement the method in their driver
    """

    def __init__(self, conf):
        pass

    def configure_healthmonitor(self, context, resource_data):
        """Checks if the Service VM is reachable.

           It does netcat to the CONFIGURATION_SERVER_PORT of the Service VM.
           Configuration agent runs inside Service VM. Once agent is up and
           reachable, Service VM is assumed to be active.

           :param context - context
           :param resource_data - data coming from orchestrator

           Returns: SUCCESS/FAILED

        """

        resource_data = self.parse.parse_data(const.HEALTHMONITOR,
                                              resource_data)
        ip = resource_data.get('mgmt_ip')
        port = str(self.port)
        command = 'nc ' + ip + ' ' + port + ' -z'
        return self._check_vm_health(command)

    def configure_interfaces(self, context, kwargs):
        return const.SUCCESS

    def clear_interfaces(self, context, kwargs):
        return const.SUCCESS

    def configure_routes(self, context, kwargs):
        return const.SUCCESS

    def clear_routes(self, context, kwargs):
        return const.SUCCESS

    def clear_healthmonitor(self, context, kwargs):
        return const.SUCCESS

    def register_agent_object_with_driver(self, name, agent_obj):
        setattr(BaseDriver, name, agent_obj)

    def _check_vm_health(self, command):
        """Ping based basic HM support provided by BaseDriver.
           Service provider can override the method implementation
           if they want to support other types.

           :param command - command to execute

           Returns: SUCCESS/FAILED
        """
        msg = ("Executing command %s for VM health check" % (command))
        LOG.debug(msg)
        try:
            subprocess.check_output(command, stderr=subprocess.STDOUT,
                                    shell=True)
        except Exception as e:
            msg = ("VM health check failed. Command '%s' execution failed."
                   " Reason=%s" % (command, e))
            LOG.debug(msg)
            return const.FAILED
        return const.SUCCESS

    def _configure_log_forwarding(self, url, mgmt_ip, port):
        """ Configures log forwarding IP address in Service VMs.

            :param url: url format that is used to invoke the Service VM API
            :param mgmt_ip: management IP of the Service VM
            :param port: port that is listened to by the Service VM agent

            Returns: SUCCESS/Error msg

        """

        url = url % (mgmt_ip, port, 'configure-rsyslog-as-client')

        log_forward_ip_address = self.conf.configurator.log_forward_ip_address
        if not log_forward_ip_address:
            msg = ("Log forwarding IP address not configured "
                   "for service at %s." % mgmt_ip)
            LOG.info(msg)
            return const.UNHANDLED

        data = dict(
            server_ip=log_forward_ip_address,
            server_port=self.conf.configurator.log_forward_port,
            log_level=self.conf.configurator.log_level)
        data = jsonutils.dumps(data)

        msg = ("Initiating POST request to configure log forwarding "
               "for service at: %r" % mgmt_ip)
        LOG.info(msg)

        try:
            resp = requests.post(url, data, timeout=self.timeout)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to service at: "
                   "%r for configuring log forwarding. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened while configuring "
                   "log forwarding for service at: %r. "
                   "ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response of configure log forward API, "
                   "invalid JSON. URL: %r. %r" % (url, str(err).capitalize()))
            LOG.error(msg)
            return msg
        if not result['status']:
            msg = ("Error configuring log forwarding for service "
                   "at %s. URL: %r. Reason: %s." %
                   (mgmt_ip, url, result['reason']))
            LOG.error(msg)
            return msg

        msg = ("Successfully configured log forwarding for "
               "service at %s." % mgmt_ip)
        LOG.info(msg)
        return const.SUCCESS



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
import os
import oslo_serialization.jsonutils as jsonutils
import pecan
PECAN_CONFIG_FILE = (os.getcwd() +
                     "/gbpservice/nfp/pecan/api/config.py")

pecan.set_config(PECAN_CONFIG_FILE, overwrite=True)

import webtest
import zlib

from neutron.tests import base
from pecan import rest

from gbpservice.nfp.pecan import constants

setattr(pecan, 'mode', constants.advanced)

from gbpservice.contrib.nfp.configurator.advanced_controller import controller
from gbpservice.nfp.pecan.api import root_controller
reload(root_controller)


class ControllerTestCase(base.BaseTestCase, rest.RestController):
    """
    This class contains all the unittest cases for REST server of configurator.

    This class tests success and failure cases for all the HTTP requests which
    are implemented in REST server. run_tests.sh file is used for running all
    the tests in this class. All the methods of this class started with test
    prefix called and on success it will print ok and on failure it will
    print the error trace.

    """
    @classmethod
    def setUpClass(cls):
        """A class method called before tests in an individual class run

        """
        rootController = root_controller.RootController()
        ControllerTestCase.app = webtest.TestApp(
                                            pecan.make_app(rootController))
        ControllerTestCase.data = {'info': {'service_type': 'firewall',
                                            'service_vendor': 'vyos',
                                            'context': {}},
                                   'config': [{'resource': 'firewall',
                                               'resource_data': {}}]
                                   }

    def test_get_notifications(self):
        """Tests HTTP get request get_notifications.

        Returns: none

        """
        with mock.patch.object(
                controller.RMQConsumer, 'pull_notifications') as mock_pn:
            response = self.app.get('/v1/nfp/get_notifications')
        mock_pn.assert_called_with()
        self.assertEqual(response.status_code, 200)

    def test_post_create_network_function_device_config(self):
        """Tests HTTP post request create_network_function_device_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.post(
                '/v1/nfp/create_network_function_device_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'create_network_function_device_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_post_create_network_function_config(self):
        """Tests HTTP post request create_network_function_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.post(
                '/v1/nfp/create_network_function_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'create_network_function_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_post_delete_network_function_device_config(self):
        """Tests HTTP post request delete_network_function_device_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.post(
                '/v1/nfp/delete_network_function_device_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'delete_network_function_device_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_post_delete_network_function_config(self):
        """Tests HTTP post request delete_network_function_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.post(
                '/v1/nfp/delete_network_function_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'delete_network_function_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_put_update_network_function_device_config(self):
        """Tests HTTP put request update_network_function_device_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.put(
                '/v1/nfp/update_network_function_device_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'update_network_function_device_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_put_update_network_function_config(self):
        """Tests HTTP put request update_network_function_config.

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            response = self.app.put(
                '/v1/nfp/update_network_function_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        rpc_mock.assert_called_with(
            'update_network_function_config', self.data)
        self.assertEqual(response.status_code, 200)

    def test_post_create_network_function_device_config_fail(self):
        """Tests failure case of HTTP post request
        create_network_function_device_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/create_network_function_device_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)

    def test_post_create_network_function_config_fail(self):
        """Tests failure case of HTTP post request
        create_network_function_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/create_network_function_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)

    def test_post_delete_network_function_device_config_fail(self):
        """Tests failure case of HTTP post request
        delete_network_function_device_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/delete_network_function_device_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)

    def test_post_delete_network_function_config_fail(self):
        """Tests failure case of HTTP post request
        delete_network_function_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/delete_network_function_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)

    def test_put_update_network_function_device_config_fail(self):
        """Tests failure case of HTTP put request
        update_network_function_device_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/update_network_function_device_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)

    def test_put_update_network_function_config_fail(self):
        """Tests failure case of HTTP put request
        update_network_function_config

        Returns: none

        """

        with mock.patch.object(
                controller.RPCClient, 'cast') as rpc_mock:
            rpc_mock.return_value = Exception
            response = self.app.post(
                '/v1/nfp/update_network_function_config',
                expect_errors=True)
            self.assertEqual(response.status_code, 400)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.common import config  # noqa

from gbpservice.neutron.services.servicechain.plugins.ncp.node_plumbers import(
    admin_owned_resources_apic_tscp as admin_tscp)
from gbpservice.neutron.tests.unit.services.servicechain.ncp import (
    test_tscp_apic_mapping as test_tscp_apic_mapping)


class AdminOwnedResourcesTscpTestCase(
        test_tscp_apic_mapping.ApicMappingStitchingPlumberGBPTestCase):

    def setUp(self):
        user = 'user'
        password = 'password'
        tenant_name = 'tenant_name',
        uri = 'http://127.0.0.1:35357/v2.0/'
        try:
            config.cfg.CONF.keystone_authtoken.username
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('username', default=user),
                'keystone_authtoken')
        else:
            config.cfg.CONF.set_override('username', user,
                                         group='keystone_authtoken')
        try:
            config.cfg.CONF.keystone_authtoken.password
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('password', default=password),
                'keystone_authtoken')
        else:
            config.cfg.CONF.set_override('password', password,
                                         group='keystone_authtoken')
        try:
            config.cfg.CONF.keystone_authtoken.project_name
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('project_name', default=tenant_name),
                'keystone_authtoken')
        else:
            config.cfg.CONF.set_override('project_name', tenant_name,
                                         group='keystone_authtoken')
        config.cfg.CONF.set_override('auth_uri', uri,
                                     group='keystone_authtoken')
        super(AdminOwnedResourcesTscpTestCase, self).setUp(
            plumber='admin_owned_resources_apic_plumber')
        admin_tscp.keyclient = mock.Mock()
        res = mock.patch('gbpservice.neutron.services.servicechain.plugins.'
                         'ncp.node_plumbers.admin_owned_resources_apic_tscp.'
                         'AdminOwnedResourcesApicTSCP.'
                         '_get_resource_owner_tenant_id').start()
        res.return_value = "1234"


class TestApicChains(AdminOwnedResourcesTscpTestCase,
                     test_tscp_apic_mapping.TestApicChains):
    pass



# Copyright (c) 2016 Cisco Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron import context
from neutron_lib import constants as q_const
from neutron_lib import exceptions as n_exc

from gbpservice.neutron.services.l3_router import l3_apic
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_apic_mapping)


TENANT = 'tenant1'
ROUTER = 'router1'
SUBNET = 'subnet1'
NETWORK = 'network1'
PORT = 'port1'
NETWORK_NAME = 'one_network'
TEST_SEGMENT1 = 'test-segment1'
FLOATINGIP = 'fip1'


# TODO(tbachman): create better test class hierarchy to inherit
class TestCiscoApicL3Plugin(test_apic_mapping.ApicMappingTestCase):
    '''Test class for the Cisco APIC L3 Plugin

       This is a set of tests specific to the Cisco APIC
       L3 plugin. It currently derives from the ApicMappingTestCase
       class so that it can inherit test infrastructure from those classes.
    '''

    def setUp(self):
        super(TestCiscoApicL3Plugin, self).setUp()

        # Some actual dicts to return
        self.subnet = {'network_id': NETWORK, 'tenant_id': TENANT}
        self.port = {'tenant_id': TENANT,
                     'network_id': NETWORK,
                     'fixed_ips': [{'subnet_id': SUBNET}],
                     'id': 'port_id'}
        self.interface_info = {'subnet': {'subnet_id': SUBNET},
                               'port': {'port_id': self.port['id']}}
        self.floatingip = {'id': FLOATINGIP,
                           'floating_network_id': NETWORK_NAME,
                           'port_id': PORT}
        self.context = context.get_admin_context()
        self.context.tenant_id = TENANT

        # Create our plugin, but mock some superclass and
        # core plugin methods
        self.plugin = l3_apic.ApicGBPL3ServicePlugin()
        self.plugin._apic_driver._notify_port_update = mock.Mock()

        self.plugin._core_plugin.get_ports = mock.Mock(
            return_value=[self.port])
        self.plugin._core_plugin.get_port = mock.Mock(return_value=self.port)
        self.plugin._core_plugin.get_subnet = mock.Mock(
            return_value = self.subnet)
        self.plugin._core_plugin.update_port_status = mock.Mock()

        # Floating IP updates to agents are mocked
        self.plugin.update_floatingip_status = mock.Mock()
        self.plugin.get_floatingip = mock.Mock(return_value=self.floatingip)

    def _check_call_list(self, expected, observed):
        for call in expected:
            self.assertTrue(call in observed,
                            msg='Call not found, expected:\n%s\nobserved:'
                                '\n%s' % (str(call), str(observed)))
            observed.remove(call)
        self.assertFalse(
            len(observed),
            msg='There are more calls than expected: %s' % str(observed))

    def _test_add_router_interface(self, interface_info):
        with mock.patch('neutron.db.l3_db.'
                        'L3_NAT_db_mixin.add_router_interface') as if_mock:
            if_mock.return_value = self.port
            port = self.plugin.add_router_interface(self.context,
                                                    ROUTER, interface_info)
            self.assertEqual(port, self.port)
            test_assert = self.plugin._core_plugin.update_port_status
            test_assert.assert_called_once_with(self.context,
                self.port['id'], q_const.PORT_STATUS_ACTIVE)

    def _test_remove_router_interface(self, interface_info):
        with mock.patch('neutron.db.l3_db.'
                        'L3_NAT_db_mixin.remove_router_interface') as if_mock:
            self.plugin.remove_router_interface(self.context, ROUTER,
                                                interface_info)
            self.assertEqual(1, if_mock.call_count)

    def test_add_router_interface_subnet(self):
        self._test_add_router_interface(self.interface_info['subnet'])

    def test_add_router_interface_port(self):
        self._test_add_router_interface(self.interface_info['port'])

    def test_remove_router_interface_subnet(self):
        self._test_remove_router_interface(self.interface_info['subnet'])

    def test_remove_router_interface_port(self):
        self._test_remove_router_interface(self.interface_info['port'])

    def test_create_router_gateway_fails(self):
        # Force _update_router_gw_info failure
        with mock.patch('neutron.db.l3_db.L3_NAT_dbonly_mixin.'
                        '_update_router_gw_info',
                        side_effect=n_exc.NeutronException):
            data = {'router': {'tenant_id': 'foo',
                'name': 'router1', 'admin_state_up': True,
                'external_gateway_info': {'network_id': 'some_uuid'}}}
            # Verify router doesn't persist on failure
            self.assertRaises(n_exc.NeutronException,
                              self.plugin.create_router, self.context, data)
            routers = self.plugin.get_routers(self.context)
            self.assertEqual(0, len(routers))

    def test_floatingip_port_notify_on_create(self):
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'create_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            # create floating-ip with mapped port
            plugin = self.plugin
            plugin.create_floatingip(self.context,
                                     {'floatingip': self.floatingip})
            plugin._apic_driver._notify_port_update.assert_called_once_with(
                PORT)

    def test_floatingip_port_notify_on_reassociate(self):
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'update_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            # associate with different port
            new_fip = {'port_id': 'port-another'}
            self.plugin.update_floatingip(self.context, FLOATINGIP,
                                          {'floatingip': new_fip})
            self._check_call_list(
                [mock.call(PORT),
                 mock.call('port-another')],
                self.plugin._apic_driver._notify_port_update.call_args_list)

    def test_floatingip_port_notify_on_disassociate(self):
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'update_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            # dissociate mapped port
            plugin = self.plugin
            plugin.update_floatingip(self.context, FLOATINGIP,
                                     {'floatingip': {}})
            plugin._apic_driver._notify_port_update.assert_any_call(
                PORT)

    def test_floatingip_port_notify_on_delete(self):
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.delete_floatingip'):
            # delete
            plugin = self.plugin
            plugin.delete_floatingip(self.context, FLOATINGIP)
            plugin._apic_driver._notify_port_update.assert_called_once_with(
                PORT)

    def test_floatingip_status(self):
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'create_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            # create floating-ip with mapped port
            fip = self.plugin.create_floatingip(self.context,
                {'floatingip': self.floatingip})
            self.plugin.update_floatingip_status.assert_called_once_with(
                mock.ANY, FLOATINGIP, q_const.FLOATINGIP_STATUS_ACTIVE)
            self.assertEqual(q_const.FLOATINGIP_STATUS_ACTIVE, fip['status'])

        # dissociate mapped-port
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'update_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            self.plugin.update_floatingip_status.reset_mock()
            self.floatingip.pop('port_id')
            fip = self.plugin.update_floatingip(self.context, FLOATINGIP,
                {'floatingip': self.floatingip})
            self.plugin.update_floatingip_status.assert_called_once_with(
                mock.ANY, FLOATINGIP, q_const.FLOATINGIP_STATUS_DOWN)
            self.assertEqual(q_const.FLOATINGIP_STATUS_DOWN, fip['status'])

        # re-associate mapped-port
        with mock.patch('neutron.db.l3_db.L3_NAT_db_mixin.'
                        'update_floatingip',
                        new=mock.Mock(return_value=self.floatingip)):
            self.plugin.update_floatingip_status.reset_mock()
            self.floatingip['port_id'] = PORT
            fip = self.plugin.update_floatingip(self.context, FLOATINGIP,
                {'floatingip': self.floatingip})
            self.plugin.update_floatingip_status.assert_called_once_with(
                mock.ANY, FLOATINGIP, q_const.FLOATINGIP_STATUS_ACTIVE)
            self.assertEqual(q_const.FLOATINGIP_STATUS_ACTIVE, fip['status'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron import manager as n_manager
from neutron.plugins.common import constants as pconst
from oslo_log import log as logging

from gbpservice._i18n import _LE
from gbpservice.neutron.services.grouppolicy.common import exceptions as gp_exc
from gbpservice.neutron.services.grouppolicy import plugin as gbp_plugin


LOG = logging.getLogger(__name__)


class SharingMixin(object):

    """Implementation of the Service Chain Plugin sharing rules.

    """

    usage_graph = {'servicechain_spec': {'nodes':
                                         'servicechain_node'},
                   'servicechain_node': {'service_profile_id':
                                         'service_profile'},
                   'servicechain_instance': {},
                   'service_profile': {},
                   }

    @property
    def gbp_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = n_manager.NeutronManager.get_service_plugins()
        gbp_plugin = plugins.get(pconst.GROUP_POLICY)
        if not gbp_plugin:
            LOG.error(_LE("No group policy service plugin found."))
            raise gp_exc.GroupPolicyDeploymentError()
        return gbp_plugin

    def _validate_shared_create(self, context, obj, identity):
        return gbp_plugin.GroupPolicyPlugin._validate_shared_create(
            self, context, obj, identity)

    def _validate_shared_update(self, context, original, updated, identity):
        self._validate_shared_create(context, updated, identity)
        if updated.get('shared') != original.get('shared'):
            context = context.elevated()
            getattr(self, '_validate_%s_unshare' % identity)(context, updated)

    def _validate_servicechain_node_unshare(self, context, obj):
        # Verify not pointed by shared SCS
        gbp_plugin.GroupPolicyPlugin._check_shared_or_different_tenant(
            context, obj, self.get_servicechain_specs, 'id',
            obj['servicechain_specs'])

    def _validate_servicechain_spec_unshare(self, context, obj):
        # Verify not pointed by shared policy actions
        gbp_plugin.GroupPolicyPlugin._check_shared_or_different_tenant(
            context, obj, self.gbp_plugin.get_policy_actions, 'action_value',
            [obj['id']])

    def _validate_service_profile_unshare(self, context, obj):
        gbp_plugin.GroupPolicyPlugin._check_shared_or_different_tenant(
            context, obj, self.get_servicechain_nodes, 'service_profile_id')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.notifiers import nova as n_nova
from novaclient import exceptions as nova_exceptions
from oslo_log import log as logging

from gbpservice._i18n import _LW


LOG = logging.getLogger(__name__)


client = None


def _get_client():
    global client
    if client is None:
        client = n_nova.Notifier().nclient
    return client


class NovaClient(object):

    def __init__(self):
        self.client = _get_client()

    def get_server(self, server_id):
        try:
            return self.client.servers.get(server_id)
        except nova_exceptions.NotFound:
            LOG.warning(_LW("Nova returned NotFound for server: %s"),
                        server_id)
        except Exception as e:
            LOG.exception(e)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa


class ServicePolicyPTGIpAddressMapping(model_base.BASEV2):
    """Service Policy to IP Address mapping DB."""

    __tablename__ = 'gpm_service_policy_ipaddress_mappings'
    service_policy_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_network_service_policies.id'),
        nullable=False, primary_key=True)
    policy_target_group = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id'),
        nullable=False, primary_key=True)
    ipaddress = sa.Column(sa.String(36))


class ServicePolicyPTGFipMapping(model_base.BASEV2):
    """Service Policy to FIP Address mapping DB."""

    __tablename__ = 'gpm_service_policy_fip_mappings'
    service_policy_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_network_service_policies.id',
                                     ondelete='CASCADE'),
        nullable=False, primary_key=True)
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id',
                                     ondelete='CASCADE'),
        nullable=False, primary_key=True)
    floatingip_id = sa.Column(sa.String(36),
                              sa.ForeignKey('floatingips.id',
                                     ondelete='CASCADE'),
                              nullable=False,
                              primary_key=True)


class PolicyTargetFloatingIPMapping(model_base.BASEV2):
    """Mapping of PolicyTarget to Floating IP."""
    __tablename__ = 'gpm_pt_floatingip_mappings'
    policy_target_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_targets.id',
                                     ondelete='CASCADE'),
        nullable=False, primary_key=True)
    floatingip_id = sa.Column(sa.String(36),
                              sa.ForeignKey('floatingips.id',
                                            ondelete='CASCADE'),
                              nullable=False,
                              primary_key=True)


class NetworkServicePolicyMappingMixin(object):

    def _set_policy_ipaddress_mapping(self, session, service_policy_id,
                                      policy_target_group, ipaddress):
        with session.begin(subtransactions=True):
            mapping = ServicePolicyPTGIpAddressMapping(
                service_policy_id=service_policy_id,
                policy_target_group=policy_target_group, ipaddress=ipaddress)
            session.add(mapping)

    def _get_ptg_policy_ipaddress_mapping(self, session, policy_target_group):
        with session.begin(subtransactions=True):
            return (session.query(ServicePolicyPTGIpAddressMapping).
                    filter_by(policy_target_group=policy_target_group).first())

    def _delete_policy_ipaddress_mapping(self, session, policy_target_group):
        with session.begin(subtransactions=True):
            ip_mapping = session.query(
                ServicePolicyPTGIpAddressMapping).filter_by(
                    policy_target_group=policy_target_group).first()
            if ip_mapping:
                session.delete(ip_mapping)

    def _set_ptg_policy_fip_mapping(self, session, service_policy_id,
                                policy_target_group_id, fip_id):
        with session.begin(subtransactions=True):
            mapping = ServicePolicyPTGFipMapping(
                service_policy_id=service_policy_id,
                policy_target_group_id=policy_target_group_id,
                floatingip_id=fip_id)
            session.add(mapping)

    def _get_ptg_policy_fip_mapping(self, session, policy_target_group_id):
        with session.begin(subtransactions=True):
            return (session.query(ServicePolicyPTGFipMapping).
                    filter_by(policy_target_group_id=policy_target_group_id).
                    all())

    def _delete_ptg_policy_fip_mapping(self, session, policy_target_group_id):
        with session.begin(subtransactions=True):
            mappings = session.query(
                ServicePolicyPTGFipMapping).filter_by(
                    policy_target_group_id=policy_target_group_id).all()
            for mapping in mappings:
                session.delete(mapping)

    def _set_pt_floating_ips_mapping(self, session, policy_target_id, fip_ids):
        with session.begin(subtransactions=True):
            for fip_id in fip_ids:
                mapping = PolicyTargetFloatingIPMapping(
                    policy_target_id=policy_target_id, floatingip_id=fip_id)
                session.add(mapping)

    def _set_pts_floating_ips_mapping(self, session, pt_fip_map):
        with session.begin(subtransactions=True):
            for policy_target_id in pt_fip_map:
                self._set_pt_floating_ips_mapping(
                    session, policy_target_id,
                    pt_fip_map[policy_target_id])

    def _get_pt_floating_ip_mapping(self, session, policy_target_id):
        with session.begin(subtransactions=True):
            return (session.query(PolicyTargetFloatingIPMapping).
                    filter_by(policy_target_id=policy_target_id).all())

    def _delete_pt_floating_ip_mapping(self, session, policy_target_id):
        with session.begin(subtransactions=True):
            fip_mappings = session.query(
                PolicyTargetFloatingIPMapping).filter_by(
                    policy_target_id=policy_target_id).all()
            for fip_mapping in fip_mappings:
                session.delete(fip_mapping)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneclient import exceptions as k_exceptions
from keystoneclient.v2_0 import client as k_client
from neutron_lib.db import model_base
from oslo_config import cfg
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
import sqlalchemy as sa

from gbpservice._i18n import _LE
from gbpservice.common import utils
from gbpservice.network.neutronv2 import local_api
from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpdb
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy.common import constants as gconst
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc
from gbpservice.neutron.services.grouppolicy.common import utils as gutils
from gbpservice.neutron.services.grouppolicy.drivers import nsp_manager
from gbpservice.neutron.services.grouppolicy import sc_notifications


LOG = logging.getLogger(__name__)
SCI_CONSUMER_NOT_AVAILABLE = 'N/A'

chain_mapping_opts = [
    cfg.StrOpt('chain_owner_user',
               help=_("Chain owner username. If set, will be used in "
                      "place of the Neutron service admin for retrieving "
                      "tenant owner information through Keystone."),
               default=''),
    cfg.StrOpt('chain_owner_password',
               help=_("Chain owner password."), default='',
               secret=True),
    cfg.StrOpt('chain_owner_tenant_name',
               help=_("Name of the Tenant that will own the service chain "
                      "instances for this driver. Leave empty for provider "
                      "owned chains."), default=''),

]

cfg.CONF.register_opts(chain_mapping_opts, "chain_mapping")
cfg.CONF.import_opt('policy_drivers',
                    'gbpservice.neutron.services.grouppolicy.config',
                    group='group_policy')


class PtgServiceChainInstanceMapping(model_base.BASEV2, model_base.HasProject):
    """Policy Target Group to ServiceChainInstance mapping DB."""

    __tablename__ = 'gpm_ptgs_servicechain_mapping'
    provider_ptg_id = sa.Column(sa.String(36),
                                sa.ForeignKey('gp_policy_target_groups.id',
                                              ondelete='CASCADE'),
                                nullable=False)
    # Consumer PTG could be an External Policy
    consumer_ptg_id = sa.Column(sa.String(36), nullable=False)
    servicechain_instance_id = sa.Column(sa.String(36),
                                         sa.ForeignKey('sc_instances.id',
                                                       ondelete='CASCADE'),
                                         primary_key=True)


class ChainMappingDriver(api.PolicyDriver, local_api.LocalAPI,
                         nsp_manager.NetworkServicePolicyMappingMixin,
                         sc_notifications.ServiceChainNotificationsMixin):
    """Resource Mapping driver for Group Policy plugin.

    This driver implements service chain semantics by mapping group
    policy resources to various service chain constructs.
    """

    @log.log_method_call
    def initialize(self):
        self._cached_agent_notifier = None
        self.chain_owner = ChainMappingDriver.chain_tenant_id(reraise=True)

    @staticmethod
    def chain_tenant_id(reraise=False):
        try:
            keystone = ChainMappingDriver.chain_tenant_keystone_client()
        except cfg.NoSuchOptError:
            return None
        if keystone:
            tenant = cfg.CONF.chain_mapping.chain_owner_tenant_name
            try:
                # Can it be retrieved directly, without a further keystone
                # call?
                tenant = keystone.tenants.find(name=tenant)
                return tenant.id
            except k_exceptions.NotFound:
                with excutils.save_and_reraise_exception(reraise=reraise):
                    LOG.error(_LE('No tenant with name %s exists.'), tenant)
            except k_exceptions.NoUniqueMatch:
                with excutils.save_and_reraise_exception(reraise=reraise):
                    LOG.error(_LE('Multiple tenants matches found for %s'),
                              tenant)

    @staticmethod
    def chain_tenant_keystone_client():
        chain_user = cfg.CONF.chain_mapping.chain_owner_user
        user, pwd, tenant, auth_url = utils.get_keystone_creds()
        user = (chain_user or user)
        pwd = (cfg.CONF.chain_mapping.chain_owner_password or
               (pwd if not chain_user else ''))

        # Tenant must be configured in the resource_mapping section, provider
        # owner will be used otherwise.
        tenant = cfg.CONF.chain_mapping.chain_owner_tenant_name

        if tenant:
            return k_client.Client(username=user, password=pwd,
                                   auth_url=auth_url)

    @log.log_method_call
    def precommit_wrapper_for_chain_mapping(self, caller_method, context):
        # This wrapper is called from both precommit and postcommit for
        # deciding ordering for execution of precommit and postcommit
        if 'precommit' in caller_method:
            context._plugin_context.commit_phase = gconst.PRE_COMMIT
            # In case of aim_mapping driver (or a similar precommit driver),
            # postcommit functionality will be executed as a part of precommit
            if gutils.is_precommit_policy_driver_configured():
                if 'delete_policy_target_group' in caller_method:
                    context.ptg_chain_map = self._get_ptg_servicechain_mapping(
                        context._plugin_context.session,
                        context.current['id'])
                    self._cleanup_redirect_action(context)
                else:
                    method = getattr(self, '_' + caller_method.replace(
                                         'precommit', 'postcommit'))
                    method(context)
        else:
            context._plugin_context.commit_phase = gconst.POST_COMMIT
            # If driver is not aim_mapping (or a similar precommit driver),
            # then postcommit functionality will  be executed by the call
            # from gbp plugin
            if not gutils.is_precommit_policy_driver_configured():
                    method = getattr(self, '_' + caller_method)
                    method(context)
            elif 'create_policy_target_group' in caller_method or (
                'create_external_policy' in caller_method):
                if hasattr(context, 'provider_context') and (
                    hasattr(context, 'servicechain_attrs')):
                    context.provider_context.commit_phase = (
                        context._plugin_context.commit_phase)
                    context.provider_context.servicechain_instance = (
                        context._plugin_context.servicechain_instance)
                    super(ChainMappingDriver,
                        self)._create_servicechain_instance(
                            context.provider_context,
                            context.servicechain_attrs)

    @log.log_method_call
    def _create_policy_target_postcommit(self, context):
        if not context._plugin._is_service_target(context._plugin_context,
                                                  context.current['id']):
            mappings = self._get_ptg_servicechain_mapping(
                context._plugin_context.session,
                provider_ptg_id=context.current['policy_target_group_id'])
            for mapping in mappings:
                chain_context = self._get_chain_admin_context(
                    context._plugin_context,
                    instance_id=mapping.servicechain_instance_id)
                self._notify_sc_plugin_pt_added(
                    chain_context, context.current,
                    mapping.servicechain_instance_id)

    @log.log_method_call
    def create_policy_target_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_target_precommit', context)

    @log.log_method_call
    def create_policy_target_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_target_postcommit', context)

    @log.log_method_call
    def delete_policy_target_precommit(self, context):
        context._is_service_target = context._plugin._is_service_target(
            context._plugin_context, context.current['id'])

    @log.log_method_call
    def delete_policy_target_postcommit(self, context):
        if not context._is_service_target:
            mappings = self._get_ptg_servicechain_mapping(
                context._plugin_context.session,
                provider_ptg_id=context.current['policy_target_group_id'])
            for mapping in mappings:
                chain_context = self._get_chain_admin_context(
                    context._plugin_context,
                    instance_id=mapping.servicechain_instance_id)
                self._notify_sc_plugin_pt_removed(
                    chain_context, context.current,
                    mapping.servicechain_instance_id)

    @log.log_method_call
    def _create_policy_target_group_postcommit(self, context):
        if (context.current['provided_policy_rule_sets'] and
                self._is_group_chainable(context, context.current)):
            self._handle_redirect_action(
                context, context.current['provided_policy_rule_sets'],
                providing_ptg=context.current)
        self._handle_prs_added(context)
        self._handle_provider_updated(context)

    @log.log_method_call
    def create_policy_target_group_precommit(self, context):
        self._validate_ptg_prss(context, context.current)
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_target_group_precommit', context)

    @log.log_method_call
    def create_policy_target_group_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_target_group_postcommit', context)

    @log.log_method_call
    def _update_policy_target_group_postcommit(self, context):
        # Update service chain instance when any ruleset is changed
        orig = context.original
        curr = context.current

        new_provided_policy_rule_sets = list(
            set(curr['provided_policy_rule_sets']) - set(
                orig['provided_policy_rule_sets']))

        # Only the ones set in context in precommit operation will be deleted
        self._cleanup_redirect_action(context)
        # If the spec is changed, then update the chain with new spec
        # If redirect is newly added, create the chain
        if (self._is_redirect_in_policy_rule_sets(
                context, new_provided_policy_rule_sets) and
                self._is_group_chainable(context, context.current)):
            self._handle_redirect_action(
                context, curr['provided_policy_rule_sets'],
                providing_ptg=context.current)
        self._handle_prs_updated(context)
        self._handle_provider_updated(context)

    @log.log_method_call
    def update_policy_target_group_precommit(self, context):
        self._validate_ptg_prss(context, context.current)
        self._stash_ptg_modified_chains(context)
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_target_group_precommit', context)

    @log.log_method_call
    def update_policy_target_group_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_target_group_postcommit', context)

    @log.log_method_call
    def get_policy_target_group_status(self, context):
        try:
            if (context.current['provided_policy_rule_sets'] and not
                context.current.get('proxied_group_id')):
                ptg_status = []
                for sci in self._get_chains_by_prs(
                            context,
                            context.current['provided_policy_rule_sets']):
                    servicechain_instance = self._get_servicechain_instance(
                                                    context._plugin_context,
                                                    sci)
                    if (servicechain_instance['provider_ptg_id'] ==
                            context.current['id']):
                        ptg_status.append(
                                {'status': servicechain_instance['status'],
                                 'status_details': servicechain_instance[
                                    'status_details']})
                # REVISIT: For now assuming there will be only
                #          one sci associated with this ptg
                if ptg_status:
                    context.current['status'] = ptg_status[0]['status']
                    context.current['status_details'] = ptg_status[0][
                            'status_details']
        except Exception:
            LOG.error(_LE('Failed to update ptg status'))

    @log.log_method_call
    def _delete_policy_target_group_postcommit(self, context):
        self._handle_prs_removed(context)

    @log.log_method_call
    def delete_policy_target_group_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_policy_target_group_precommit', context)

    @log.log_method_call
    def delete_policy_target_group_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_policy_target_group_postcommit', context)

    @log.log_method_call
    def _update_policy_classifier_postcommit(self, context):
        self._handle_classifier_update_notification(context)

    @log.log_method_call
    def update_policy_classifier_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_classifier_precommit', context)

    @log.log_method_call
    def update_policy_classifier_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_classifier_postcommit', context)

    @log.log_method_call
    def create_policy_action_precommit(self, context):
        spec_id = context.current['action_value']
        if spec_id:
            specs = self._get_servicechain_specs(
                context._plugin_context, filters={'id': [spec_id]})
            for spec in specs:
                if not spec.get('shared', False):
                    self._reject_shared(context.current, 'policy_action')

    @log.log_method_call
    def _update_policy_action_postcommit(self, context):
        self._handle_redirect_spec_id_update(context)

    @log.log_method_call
    def update_policy_action_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_action_precommit', context)

    @log.log_method_call
    def update_policy_action_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_action_postcommit', context)

    @log.log_method_call
    def create_policy_rule_precommit(self, context):
        self._reject_multiple_redirects_in_rule(context)

    @log.log_method_call
    def _update_policy_rule_postcommit(self, context):
        old_classifier_id = context.original['policy_classifier_id']
        new_classifier_id = context.current['policy_classifier_id']
        old_action_set = set(context.current['policy_actions'])
        new_action_set = set(context.original['policy_actions'])
        if (old_classifier_id != new_classifier_id or
                old_action_set != new_action_set):
            policy_rule_sets = (
                context._plugin._get_policy_rule_policy_rule_sets(
                    context._plugin_context, context.current['id']))

            old_redirect_policy_actions = context._plugin.get_policy_actions(
                        context._plugin_context,
                        filters={'id': context.original['policy_actions'],
                                 'action_type': [gconst.GP_ACTION_REDIRECT]})
            new_redirect_policy_actions = context._plugin.get_policy_actions(
                        context._plugin_context,
                        filters={'id': context.current['policy_actions'],
                                 'action_type': [gconst.GP_ACTION_REDIRECT]})
            if old_redirect_policy_actions or new_redirect_policy_actions:
                self._handle_redirect_action(context, policy_rule_sets)

    @log.log_method_call
    def update_policy_rule_precommit(self, context):
        self._reject_multiple_redirects_in_rule(context)
        old_redirect = self._get_redirect_action(context, context.original)
        new_redirect = self._get_redirect_action(context, context.current)
        if not old_redirect and new_redirect:
            # If redirect action is added, check that there's no contract that
            # already has a redirect action
            for prs in context._plugin.get_policy_rule_sets(
                    context._plugin_context,
                    {'id': context.current['policy_rule_sets']}):
                # Make sure the PRS can have a new redirect action
                self._validate_new_prs_redirect(context, prs)
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_rule_precommit', context)

    @log.log_method_call
    def update_policy_rule_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_rule_postcommit', context)

    @log.log_method_call
    def _create_policy_rule_set_postcommit(self, context):
        if context.current['child_policy_rule_sets']:
            self._handle_redirect_action(
                context, context.current['child_policy_rule_sets'])

    @log.log_method_call
    def create_policy_rule_set_precommit(self, context):
        self._reject_multiple_redirects_in_prs(context)
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_rule_set_precommit', context)

    @log.log_method_call
    def create_policy_rule_set_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_policy_rule_set_postcommit', context)

    @log.log_method_call
    def _update_policy_rule_set_postcommit(self, context):
        if self._is_redirect_rule_updated(context):
            # Handle any Redirects from the current Policy Rule Set
            self._handle_redirect_action(context, [context.current['id']])

        # Handle Update/Delete of Redirects for any child Rule Sets
        if (set(context.original['child_policy_rule_sets']) !=
                set(context.current['child_policy_rule_sets'])):
            if context.original['child_policy_rule_sets']:
                self._handle_redirect_action(
                    context, context.original['child_policy_rule_sets'])
            if context.current['child_policy_rule_sets']:
                self._handle_redirect_action(
                    context, context.current['child_policy_rule_sets'])

    @log.log_method_call
    def update_policy_rule_set_precommit(self, context):
        self._reject_multiple_redirects_in_prs(context)
        # If a redirect action is added (from 0 to one) we have to validate
        # the providing and consuming PTGs. Not needed at creation time since
        # no PTG could be possibly providing or consuming it
        old_red_count = self._multiple_pr_redirect_action_number(
            context._plugin_context.session, context.original['policy_rules'])
        new_red_count = self._multiple_pr_redirect_action_number(
            context._plugin_context.session, context.current['policy_rules'])
        if new_red_count > old_red_count:
            self._validate_new_prs_redirect(context, context.current)
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_rule_set_precommit', context)

    @log.log_method_call
    def update_policy_rule_set_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_policy_rule_set_postcommit', context)

    @log.log_method_call
    def _delete_policy_rule_set_postcommit(self, context):
        if context.current['child_policy_rule_sets']:
            self._handle_redirect_action(
                context, context.current['child_policy_rule_sets'])

    @log.log_method_call
    def delete_policy_rule_set_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_policy_rule_set_precommit', context)

    @log.log_method_call
    def delete_policy_rule_set_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_policy_rule_set_postcommit', context)

    @log.log_method_call
    def _create_external_policy_postcommit(self, context):
        self._handle_prs_added(context)

    @log.log_method_call
    def create_external_policy_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_external_policy_precommit', context)

    @log.log_method_call
    def create_external_policy_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'create_external_policy_postcommit', context)

    @log.log_method_call
    def _update_external_policy_postcommit(self, context):
        self._handle_prs_updated(context)

    @log.log_method_call
    def update_external_policy_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_external_policy_precommit', context)

    @log.log_method_call
    def update_external_policy_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'update_external_policy_postcommit', context)

    @log.log_method_call
    def _delete_external_policy_postcommit(self, context):
        self._handle_prs_removed(context)

    @log.log_method_call
    def delete_external_policy_precommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_external_policy_precommit', context)

    @log.log_method_call
    def delete_external_policy_postcommit(self, context):
        self.precommit_wrapper_for_chain_mapping(
            'delete_external_policy_postcommit', context)

    def _is_redirect_rule_updated(self, context):
        if (not context.original['child_policy_rule_sets']) and (
            not context.current['child_policy_rule_sets']):
            '''
            This method checks if a rule with a REDIRECT action is added to or
            removed from the PRS. If no REDIRECT action has been added or
            removed, then the processing for REDIRECT does not have to be
            performed and existing service chain instances should be
            left untouched.
            '''
            old_redirect_count = self._multiple_pr_redirect_action_number(
                context._plugin_context.session,
                context.original['policy_rules'])
            new_redirect_count = self._multiple_pr_redirect_action_number(
                context._plugin_context.session,
                context.current['policy_rules'])
            if (new_redirect_count == 0) and (old_redirect_count == 0):
                return False
            elif (new_redirect_count != old_redirect_count):
                return True
            else:
                original_policy_rules = context.original['policy_rules']
                current_policy_rules = context.current['policy_rules']
                policy_rules = context._plugin.get_policy_rules(
                    context._plugin_context,
                    filters={'id': original_policy_rules})
                for policy_rule in policy_rules:
                    original_redirect_policy_action = (
                        self._get_redirect_action(
                             context, policy_rule))
                    if (original_redirect_policy_action) and (
                        policy_rule['id'] not in current_policy_rules):
                        return True
                policy_rules = context._plugin.get_policy_rules(
                    context._plugin_context,
                    filters={'id': current_policy_rules})
                for policy_rule in policy_rules:
                    current_redirect_policy_action = self._get_redirect_action(
                        context, policy_rule)
                    if (current_redirect_policy_action) and (
                        policy_rule['id'] not in original_policy_rules):
                        return True
                return False
        return True

    def _handle_prs_added(self, context):
        # Expecting either a PTG or EP context
        if context.current['consumed_policy_rule_sets']:
            for sci in self._get_chains_by_prs(
                    context, context.current['consumed_policy_rule_sets']):
                chain_context = self._get_chain_admin_context(
                    context._plugin_context, instance_id=sci)
                self._notify_sc_consumer_added(
                    chain_context, context.current, sci)

    def _handle_prs_removed(self, context):
        # Expecting either a PTG or EP context
        if context.current['consumed_policy_rule_sets']:
            for sci in self._get_chains_by_prs(
                    context, context.current['consumed_policy_rule_sets']):
                chain_context = self._get_chain_admin_context(
                    context._plugin_context, instance_id=sci)
                self._notify_sc_consumer_removed(
                    chain_context, context.current, sci)

    def _handle_prs_updated(self, context):
        # Expecting either a PTG or EP context
        if (context.current['consumed_policy_rule_sets'] !=
                context.original['consumed_policy_rule_sets']):
            added, removed = utils.set_difference(
                context.current['consumed_policy_rule_sets'],
                context.original['consumed_policy_rule_sets'])
            if removed:
                for sci in self._get_chains_by_prs(context, removed):
                    chain_context = self._get_chain_admin_context(
                        context._plugin_context, instance_id=sci)
                    self._notify_sc_consumer_removed(
                        chain_context, context.current, sci)
            if added:
                for sci in self._get_chains_by_prs(context, added):
                    chain_context = self._get_chain_admin_context(
                        context._plugin_context, instance_id=sci)
                    self._notify_sc_consumer_removed(
                        chain_context, context.current, sci)

    def _handle_redirect_spec_id_update(self, context):
        if (context.current['action_type'] != gconst.GP_ACTION_REDIRECT
            or context.current['action_value'] ==
            context.original['action_value']):
            return

        spec = self._servicechain_plugin._get_servicechain_spec(
                    context._plugin_context, context.original['action_value'])
        for servicechain_instance in spec.instances:
            sc_instance_id = servicechain_instance.servicechain_instance_id
            sc_instance = self._servicechain_plugin.get_servicechain_instance(
                    context._plugin_context, sc_instance_id)
            old_specs = sc_instance['servicechain_specs']
            # Use the parent/child redirect spec as it is. Only replace the
            # current one
            new_specs = [context.current['action_value'] if
                         x == context.original['action_value'] else
                         x for x in old_specs]
            self._update_servicechain_instance(
                context._plugin_context,
                servicechain_instance.servicechain_instance_id,
                sc_specs=new_specs)

    def _update_servicechain_instance(self, plugin_context, sc_instance_id,
                                      classifier_id=None, sc_specs=None):
        sc_instance_update_data = {}
        if sc_specs:
            sc_instance_update_data.update({'servicechain_specs': sc_specs})
        if classifier_id:
            sc_instance_update_data.update({'classifier_id': classifier_id})
        super(ChainMappingDriver, self)._update_servicechain_instance(
            self._get_chain_admin_context(
                plugin_context, instance_id=sc_instance_id),
            sc_instance_id, sc_instance_update_data)

    # This method would either update an existing chain instance, or creates a
    # new chain instance or delete the existing instance. In case of updates,
    # the parameters that can be updated are service chain spec and
    # classifier ID.
    def _handle_redirect_action(self, context, policy_rule_set_ids,
                                providing_ptg=None):
        policy_rule_sets = context._plugin.get_policy_rule_sets(
            context._plugin_context, filters={'id': policy_rule_set_ids})
        for policy_rule_set in policy_rule_sets:
            if providing_ptg:
                ptgs_providing_prs = [providing_ptg]
            else:
                if not policy_rule_set['providing_policy_target_groups']:
                    continue
                ptgs_providing_prs = context._plugin.get_policy_target_groups(
                    context._plugin_context.elevated(),
                    {'id': policy_rule_set['providing_policy_target_groups']})
            parent_classifier_id = None
            parent_spec_id = None
            if policy_rule_set['parent_id']:
                parent = context._plugin.get_policy_rule_set(
                    context._plugin_context, policy_rule_set['parent_id'])
                policy_rules = context._plugin.get_policy_rules(
                    context._plugin_context,
                    filters={'id': parent['policy_rules']})
                for policy_rule in policy_rules:
                    policy_actions = context._plugin.get_policy_actions(
                        context._plugin_context,
                        filters={'id': policy_rule["policy_actions"],
                                 'action_type': [gconst.GP_ACTION_REDIRECT]})
                    if policy_actions:
                        parent_spec_id = policy_actions[0].get("action_value")
                        parent_classifier_id = policy_rule.get(
                            "policy_classifier_id")
                        break  # only one redirect action is supported
            policy_rules = context._plugin.get_policy_rules(
                    context._plugin_context,
                    filters={'id': policy_rule_set['policy_rules']})
            for policy_rule in policy_rules:
                hierarchial_classifier_mismatch = False
                classifier_id = policy_rule.get("policy_classifier_id")
                if parent_classifier_id and (parent_classifier_id !=
                                             classifier_id):
                    hierarchial_classifier_mismatch = True
                policy_actions = context._plugin.get_policy_actions(
                        context._plugin_context,
                        filters={'id': policy_rule.get("policy_actions"),
                                 'action_type': [gconst.GP_ACTION_REDIRECT]})
                # Only one Redirect action per PRS. The chain may belong to
                # another PRS in which case the chain should not be deleted
                if (self._is_redirect_in_policy_rule_sets(
                    context, policy_rule_set_ids) and not policy_actions):
                    continue
                spec_id = (policy_actions and policy_actions[0]['action_value']
                           or None)
                for ptg_providing_prs in ptgs_providing_prs:
                    # REVISIT(Magesh): There are concurrency issues here with
                    # concurrent updates to the same PRS, Policy Rule or Action
                    # value
                    if self._is_group_chainable(context, ptg_providing_prs):
                        self._create_or_update_chain(
                            context, ptg_providing_prs['id'],
                            SCI_CONSUMER_NOT_AVAILABLE, spec_id,
                            parent_spec_id, classifier_id,
                            hierarchial_classifier_mismatch,
                            policy_rule_set)

    def _create_or_update_chain(self, context, provider, consumer, spec_id,
                                parent_spec_id, classifier_id,
                                hierarchial_classifier_mismatch, prs_id):
        ptg_chain_map = self._get_ptg_servicechain_mapping(
            context._plugin_context.session, provider)
        if ptg_chain_map:
            if hierarchial_classifier_mismatch or not spec_id:
                ctx = self._get_chain_admin_context(
                    context._plugin_context,
                    tenant_id=ptg_chain_map[0].tenant_id)
                self._delete_servicechain_instance(
                    ctx, ptg_chain_map[0].servicechain_instance_id)
            else:
                sc_specs = [spec_id]
                if parent_spec_id:
                    sc_specs.insert(0, parent_spec_id)
                # One chain per providing PTG
                self._update_servicechain_instance(
                    context._plugin_context,
                    ptg_chain_map[0].servicechain_instance_id,
                    classifier_id=classifier_id,
                    sc_specs=sc_specs)
        elif spec_id and not hierarchial_classifier_mismatch:
            self._create_servicechain_instance(
                context, spec_id, parent_spec_id, provider,
                SCI_CONSUMER_NOT_AVAILABLE, classifier_id, prs_id)

    def _cleanup_redirect_action(self, context):
        for ptg_chain in context.ptg_chain_map:
            ctx = self._get_chain_admin_context(context._plugin_context,
                                                tenant_id=ptg_chain.tenant_id)
            self._delete_servicechain_instance(
                ctx, ptg_chain.servicechain_instance_id)

    def _create_servicechain_instance(self, context, servicechain_spec,
                                      parent_servicechain_spec,
                                      provider_ptg_id, consumer_ptg_id,
                                      classifier_id, policy_rule_set):
        sc_spec = [servicechain_spec]
        if parent_servicechain_spec:
            sc_spec.insert(0, parent_servicechain_spec)
        config_param_values = {}
        provider_ptg = context._plugin.get_policy_target_group(
            utils.admin_context(context._plugin_context), provider_ptg_id)
        p_ctx = self._get_chain_admin_context(
            context._plugin_context,
            provider_tenant_id=provider_ptg['tenant_id'])
        session = context._plugin_context.session
        network_service_policy_id = provider_ptg.get(
            "network_service_policy_id")
        if network_service_policy_id:
            nsp = context._plugin.get_network_service_policy(
                p_ctx, network_service_policy_id)
            service_params = nsp.get("network_service_params")
            for service_parameter in service_params:
                param_type = service_parameter.get("type")
                param_value = service_parameter.get("value")
                if param_type == "ip_single" and param_value == "self_subnet":
                    key = service_parameter.get("name")
                    servicepolicy_ptg_ip_map = (
                        self._get_ptg_policy_ipaddress_mapping(
                            session, provider_ptg_id))
                    servicepolicy_ip = servicepolicy_ptg_ip_map.get(
                                                        "ipaddress")
                    config_param_values[key] = servicepolicy_ip
                elif param_type == "ip_single" and param_value == "nat_pool":
                    key = service_parameter.get("name")
                    fip_maps = (
                        self._get_ptg_policy_fip_mapping(
                            context._plugin_context.session,
                            provider_ptg_id))
                    servicepolicy_fip_ids = []
                    for fip_map in fip_maps:
                        servicepolicy_fip_ids.append(fip_map.floatingip_id)
                    config_param_values[key] = servicepolicy_fip_ids
        name = 'gbp_%s_%s' % (policy_rule_set['name'], provider_ptg['name'])

        attrs = {'tenant_id': p_ctx.tenant,
                 'name': name,
                 'description': "",
                 'servicechain_specs': sc_spec,
                 'provider_ptg_id': provider_ptg_id,
                 'consumer_ptg_id': SCI_CONSUMER_NOT_AVAILABLE,
                 'management_ptg_id': None,
                 'classifier_id': classifier_id,
                 'config_param_values': jsonutils.dumps(config_param_values)}
        context.provider_context = p_ctx
        context.servicechain_attrs = attrs
        sc_instance = super(
            ChainMappingDriver, self)._create_servicechain_instance(
                p_ctx, attrs)
        context._plugin_context.servicechain_instance = sc_instance
        self._set_ptg_servicechain_instance_mapping(
            session, provider_ptg_id, SCI_CONSUMER_NOT_AVAILABLE,
            sc_instance['id'], p_ctx.tenant)
        return sc_instance

    def _set_ptg_servicechain_instance_mapping(self, session, provider_ptg_id,
                                               consumer_ptg_id,
                                               servicechain_instance_id,
                                               provider_tenant_id):
        with session.begin(subtransactions=True):
            mapping = PtgServiceChainInstanceMapping(
                provider_ptg_id=provider_ptg_id,
                consumer_ptg_id=consumer_ptg_id,
                servicechain_instance_id=servicechain_instance_id,
                tenant_id=provider_tenant_id)
            session.add(mapping)

    def _get_ptg_servicechain_mapping(self, session, provider_ptg_id=None,
                                      consumer_ptg_id=None, tenant_id=None,
                                      servicechain_instance_id=None,
                                      provider_ptg_ids=None):
        with session.begin(subtransactions=True):
            query = session.query(PtgServiceChainInstanceMapping)
            if provider_ptg_id:
                query = query.filter_by(provider_ptg_id=provider_ptg_id)
            elif provider_ptg_ids:
                query = query.filter(
                    PtgServiceChainInstanceMapping.provider_ptg_id.in_(
                        list(provider_ptg_ids)))
            if consumer_ptg_id:
                query = query.filter_by(consumer_ptg_id=consumer_ptg_id)
            if servicechain_instance_id:
                query = query.filter_by(
                    servicechain_instance_id=servicechain_instance_id)
            if tenant_id:
                query = query.filter_by(consumer_ptg_id=tenant_id)
            all = query.all()
            return [utils.DictClass([('provider_ptg_id', x.provider_ptg_id),
                                     ('consumer_ptg_id', x.consumer_ptg_id),
                                     ('servicechain_instance_id',
                                      x.servicechain_instance_id),
                                     ('tenant_id', x.tenant_id)])
                    for x in all]

    def _get_chain_admin_context(self, plugin_context, tenant_id=None,
                                 provider_tenant_id=None, instance_id=None):
        ctx = plugin_context.elevated()
        # REVISIT(Ivar): Any particular implication when a provider owned PT
        # exist in the consumer PTG? Especially when the consumer PTG belongs
        # to another tenant? We may want to consider a strong convention
        # for reference plumbers to absolutely avoid this kind of inter tenant
        # object creation when the owner is the provider (in which case, the
        # context can as well be a normal context without admin capabilities).
        ctx.tenant_id = None
        if instance_id:
            cmap = self._get_ptg_servicechain_mapping(
                ctx.session, servicechain_instance_id=instance_id)
            if cmap:
                ctx.tenant_id = cmap[0].tenant_id
        if not self.chain_owner:
            self.chain_owner = ChainMappingDriver.chain_tenant_id(reraise=True)
        if not ctx.tenant_id:
            ctx.tenant_id = tenant_id or self.chain_owner or provider_tenant_id
        if self.chain_owner == ctx.tenant_id:
            ctx.auth_token = self.chain_tenant_keystone_client().get_token(
                self.chain_owner)
        return ctx

    def _is_redirect_in_policy_rule_sets(self, context, policy_rule_sets):
        policy_rule_ids = []
        for prs in context._plugin.get_policy_rule_sets(
                context._plugin_context, filters={'id': policy_rule_sets}):
            policy_rule_ids.extend(prs['policy_rules'])
        for rule in context._plugin.get_policy_rules(
                context._plugin_context, filters={'id': policy_rule_ids}):
            redirect_actions = context._plugin.get_policy_actions(
                        context._plugin_context,
                        filters={'id': rule["policy_actions"],
                                 'action_type': [gconst.GP_ACTION_REDIRECT]})
            if redirect_actions:
                return True
        return False

    def _get_redirect_action(self, context, policy_rule):
        for action in context._plugin.get_policy_actions(
                context._plugin_context,
                filters={'id': policy_rule['policy_actions']}):
            if action['action_type'] == gconst.GP_ACTION_REDIRECT:
                return action

    def _validate_new_prs_redirect(self, context, prs):
        if self._prss_redirect_rules(context._plugin_context.session,
                                     [prs['id']]) > 1:
            raise exc.MultipleRedirectActionsNotSupportedForPRS()
        for ptg in context._plugin.get_policy_target_groups(
                context._plugin_context,
                {'id': prs['providing_policy_target_groups']}):
            self._validate_ptg_prss(context, ptg)

    def _prss_redirect_rules(self, session, prs_ids):
        if len(prs_ids) == 0:
            # No result will be found in this case
            return 0
        query = (session.query(gpdb.gpdb.PolicyAction).
                 join(gpdb.gpdb.PolicyRuleActionAssociation).
                 join(gpdb.gpdb.PolicyRule).
                 join(gpdb.gpdb.PRSToPRAssociation).
                 filter(
                 gpdb.gpdb.PRSToPRAssociation.policy_rule_set_id.in_(prs_ids)).
                 filter(gpdb.gpdb.PolicyAction.action_type ==
                        gconst.GP_ACTION_REDIRECT))
        return query.count()

    def _multiple_pr_redirect_action_number(self, session, pr_ids):
        # Given a set of rules, gives the total number of redirect actions
        # found
        if len(pr_ids) == 0:
            # No result will be found in this case
            return 0
        return (session.query(gpdb.gpdb.PolicyAction).
                join(gpdb.gpdb.PolicyRuleActionAssociation).
                filter(
                gpdb.gpdb.PolicyRuleActionAssociation.policy_rule_id.in_(
                    pr_ids)).
                filter(gpdb.gpdb.PolicyAction.action_type ==
                       gconst.GP_ACTION_REDIRECT)).count()

    def _reject_shared(self, object, type):
        if object.get('shared'):
            raise exc.InvalidSharedResource(type=type,
                                            driver='chain_mapping')

    def _reject_multiple_redirects_in_rule(self, context):
        policy_actions = context._plugin.get_policy_actions(
                context._plugin_context,
                filters={'id': context.current['policy_actions'],
                         'action_type': [gconst.GP_ACTION_REDIRECT]})
        if len(policy_actions) > 1:
            raise exc.MultipleRedirectActionsNotSupportedForRule()

    def _reject_multiple_redirects_in_prs(self, context):
        policy_rules = context._plugin.get_policy_rules(
                context._plugin_context,
                filters={'id': context.current['policy_rules']})
        redirect_actions_list = []
        for policy_rule in policy_rules:
            policy_actions = context._plugin.get_policy_actions(
                    context._plugin_context,
                    filters={'id': policy_rule['policy_actions'],
                             'action_type': [gconst.GP_ACTION_REDIRECT]})
            redirect_actions_list.extend(policy_actions)
        if len(redirect_actions_list) > 1:
            raise exc.MultipleRedirectActionsNotSupportedForPRS()

    def _validate_ptg_prss(self, context, ptg):
        # If the PTG is providing a redirect PRS, it can't provide any more
        # redirect rules
        if self._is_group_chainable(context, ptg):
            if self._prss_redirect_rules(context._plugin_context.session,
                                         ptg['provided_policy_rule_sets']) > 1:
                    raise exc.PTGAlreadyProvidingRedirectPRS(ptg_id=ptg['id'])

    def _handle_classifier_update_notification(self, context):
        # Invoke Service chain update notify hook if protocol or port or
        # direction is updated. The SC side will have to reclassify the chain
        # and update the traffic steering programming
        if (context.original['port_range'] != context.current['port_range'] or
            context.original['protocol'] != context.current['protocol'] or
            context.original['direction'] != context.current['direction']):
            sc_instances = (
                self._servicechain_plugin.get_servicechain_instances(
                    context._plugin_context.elevated(),
                    filters={'classifier_id': [context.current['id']]}))
            for sc_instance in sc_instances:
                cmap = self._get_ptg_servicechain_mapping(
                    context._plugin_context.session,
                    servicechain_instance_id=sc_instance['id'])
                ctx = self._get_chain_admin_context(context._plugin_context,
                                                    cmap[0].tenant_id)
                self._servicechain_plugin.notify_chain_parameters_updated(
                    ctx, sc_instance['id'])

    def _stash_ptg_modified_chains(self, context):
        #Update service chain instance when any ruleset is changed
        orig_provided_policy_rule_sets = context.original[
            'provided_policy_rule_sets']
        curr_provided_policy_rule_sets = context.current[
            'provided_policy_rule_sets']

        removed_provided_prs = (set(orig_provided_policy_rule_sets) -
                                set(curr_provided_policy_rule_sets))
        added_provided_prs = (set(curr_provided_policy_rule_sets) -
                              set(orig_provided_policy_rule_sets))
        context.ptg_chain_map = []
        # If the Redirect is removed, delete the chain. If the spec is
        # changed, then update the existing instance with new spec
        if (self._is_redirect_in_policy_rule_sets(
                context, removed_provided_prs) and not
            self._is_redirect_in_policy_rule_sets(
                context, added_provided_prs)):
            context.ptg_chain_map += self._get_ptg_servicechain_mapping(
                context._plugin_context.session, context.current['id'])

    def _get_chains_by_prs(self, context, prs_ids):
        # REVISIT(ivar): only works under the assumption that only -one- chain
        # can be provided by a given group. A more direct way of retrieving
        # this info must be implemented before we drop this limitation
        result = set()
        for prs in self._get_policy_rule_sets(
                context._plugin_context.elevated(), {'id': prs_ids}):
            if prs['providing_policy_target_groups']:
                result |= set(
                    [x.servicechain_instance_id for x in
                     self._get_ptg_servicechain_mapping(
                         context._plugin_context.session,
                         provider_ptg_ids=prs[
                             'providing_policy_target_groups'])])
        return result

    def _is_group_chainable(self, context, group):
        """Determines whether a group should trigger a chain.

        Non chainable groups:
        - Proxy groups;

        :param context:
        :param group:
        :return: boolean
        """
        return (not group.get('proxied_group_id') and
                group.get('enforce_service_chains', True) is True)

    def _get_ptg_provided_chain_id(self, context, ptg):
        try:
            return [x.servicechain_instance_id for x in
                    self._get_ptg_servicechain_mapping(
                        context._plugin_context.session,
                        provider_ptg_ids=[ptg['id']])][0]
        except IndexError:
            return None

    def _handle_provider_updated(self, context):
        sci = self._get_ptg_provided_chain_id(context, context.current)
        if sci:
            chain_context = self._get_chain_admin_context(
                context._plugin_context, instance_id=sci)
            self._notify_ptg_updated(chain_context, context.original,
                                     context.current, sci)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from gbpservice.nfp.pecan import constants


class DecideConfigurator(pecan.commands.serve.ServeCommand):
    ''' decides the type of configurtor to be used
        like base_configurator or reference_configurator
    '''
    arguments = pecan.commands.serve.ServeCommand.arguments + ({
        'name': '--mode',
        'help': 'decides the type of configurtor to be used',
        'choices': constants.modes,
    },)

    def run(self, args):
        setattr(pecan, 'mode', args.mode)
        super(DecideConfigurator, self).run(args)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.extensions import l3
from neutron_lib.api import converters

from gbpservice.neutron.extensions import cisco_apic

ALIAS = 'cisco-apic-l3'

EXTERNAL_PROVIDED_CONTRACTS = 'apic:external_provided_contracts'
EXTERNAL_CONSUMED_CONTRACTS = 'apic:external_consumed_contracts'

CONTRACT = 'Contract'
CONTRACT_SUBJECT = 'ContractSubject'
UNSCOPED_VRF = 'no_scope-VRF'
SCOPED_VRF = 'as_%s-VRF'

EXT_GW_ATTRIBUTES = {
    EXTERNAL_PROVIDED_CONTRACTS: {
        # Additional contracts provided by external network
        'allow_put': True, 'allow_post': True,
        'is_visible': True, 'default': None,
        'convert_to': converters.convert_none_to_empty_list,
        'validate': {'type:list_of_unique_strings': None},
    },
    EXTERNAL_CONSUMED_CONTRACTS: {
        # Additional contracts consumed by external network
        'allow_put': True, 'allow_post': True,
        'is_visible': True, 'default': None,
        'convert_to': converters.convert_none_to_empty_list,
        'validate': {'type:list_of_unique_strings': None},
    }
}

EXTENDED_ATTRIBUTES_2_0 = {
    l3.ROUTERS: dict(cisco_apic.APIC_ATTRIBUTES.items() +
                     EXT_GW_ATTRIBUTES.items())
}

# Pass this key with the value True in the interface_info parameter to
# add_router_interface to override validation that normally disallows
# topologies where the same network has multiple subnets connected to
# multiple routers, which would result in unintended routing. This is
# intended only for use in the aim_mapping GBP policy driver to get
# from one valid topology to another within a transaction, and is not
# supported for any other use. It may be removed in a future release
# without any deprecation notice.
OVERRIDE_NETWORK_ROUTING_TOPOLOGY_VALIDATION = (
    'apic_aim_override_network_routing_topology_validation')


class Cisco_apic_l3(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Cisco APIC L3"

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return ("Extension exposing mapping of Neutron L3 resources to Cisco "
                "APIC constructs")

    @classmethod
    def get_updated(cls):
        return "2016-09-06T12:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}
